/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckclient;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Robot;
import java.awt.Toolkit;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import util.Logging;
import paddlepuckprotocol.PaddlePuckClient;
import com.threed.jpct.*;
import paddlepuckGUI.IPaddlePuckClientView;
import util.MyMath.Point2D;

/**
 * This class implements the client side view of a PaddlePuckGameState.
 * (uses JPCT software renderer)
 */
public class PaddlePuckClientView extends Canvas implements IPaddlePuckClientView
{
  private Logging logging = null;
  private PaddlePuckClient ownerClient = null;
  private Frame tlFrame = null; // top level frame needed for mouse control
  
  // for detecting relative mouse movement;
  private Cursor noCursor = null;
  private Robot robot = null;
  private volatile boolean mouseCaptured = false;
  
  // for rendering the gamestate
  private volatile boolean skipPaint = false;
  private FrameBuffer fb = null;
  private World world = null;
  private Object3D paddle[] = null;
  private Object3D puck = null;
  private Object3D puckShadow = null;
  
  /**
   * constructs the PaddlePuckClientView
   * @param logging the logging object to use for logging messages
   * @param tlFrame must be a reference to the top level frame which will include this view (needed for mouse control)
   * @param ownerClient must be a reference to the client whose PaddlePuckGameState will be displayed
   * @param width the desired width of the view
   * @param height the desired height of the view
   * @param antialias set to true to enable multisampling in renderer
   * @throws Exception as thrown by Robot constructor
   */
  public PaddlePuckClientView(Logging logging, Frame tlFrame, PaddlePuckClient ownerClient, int width, int height, boolean antialias) throws Exception
  {
    this.logging = logging;
    this.tlFrame = tlFrame;
    this.ownerClient = ownerClient;
    
    this.setBackground(Color.BLACK);
    this.setForeground(Color.WHITE);
    
    this.setSize(width, height);
    
    // init invisible cursor and robot
    Toolkit tk = Toolkit.getDefaultToolkit();
    Image img = tk.createImage("");
    Point hotspot = new Point(0,0);
    noCursor = tk.createCustomCursor(img, hotspot, "noCursor");
    robot = new Robot();
    
    // init jpct rendering objects
    world = new World();
    world.setAmbientLight(100, 100, 100);
    world.addLight(new SimpleVector(0f, -430f, 0f), Color.WHITE);
    TextureManager.getInstance().addTexture("texHome0", new Texture(2, 2, new Color(0,128,0)));
    TextureManager.getInstance().addTexture("texHome1", new Texture(2, 2, new Color(0,128,128)));
    TextureManager.getInstance().addTexture("texPaddle", new Texture(2, 2, new Color(128,128,255)));
    TextureManager.getInstance().addTexture("texPuck", new Texture(2, 2, new Color(64,64,64)));
    TextureManager.getInstance().addTexture("texTable", new Texture(2, 2, new Color(128,128,0)));
    TextureManager.getInstance().addTexture("texBorder", new Texture(2, 2, new Color(128, 64, 0)));
    
    // add table and homezones
    world.addObject(this.createTableCenterPiece("texTable"));
    world.addObject(this.createTableBorder("l", "texBorder"));
    world.addObject(this.createTableBorder("r", "texBorder"));
    world.addObject(this.createTableHomeZone(0, "texHome0"));
    world.addObject(this.createTableHomeZone(1, "texHome1"));
    
    // add paddles
    paddle = new Object3D[2];
    paddle[0] = this.createPaddleObject("texPaddle");
    paddle[1] = this.createPaddleObject("texPaddle");
    world.addObject(paddle[0]);
    world.addObject(paddle[1]);
    
    // add puck
    puck = this.createPuckObject(false, 20, "texPuck");
    puckShadow = this.createPuckObject(true, 20, "texPuck");
    world.addObject(puck);
    world.addObject(puckShadow);
    
    // set camera
    this.SetCameraToPlayer(0);
    
    fb = new FrameBuffer(width,height, antialias ? FrameBuffer.SAMPLINGMODE_OGSS : FrameBuffer.SAMPLINGMODE_NORMAL);
    // fb.enableRenderer(IRenderer.RENDERER_SOFTWARE); // unnecessary: the software renderer is already enabled by default and this line would cause it to get disposed and replaced by a new instance of it
    
    // attach mouse and mousemotion listeners (for paddle control)
    this.addMouseListener(new MouseAdapter() {
      @Override
      public void mouseReleased(MouseEvent e)
      {
        mouseClickedHandler(e);
      }
    });
    
    this.addMouseMotionListener(new MouseMotionAdapter() {
      @Override
      public void mouseMoved(MouseEvent e)
      {
        mouseMovedHandler(e);
      }
    });
  }
  
  /**
   * sets the camera to the respective players side of the table
   * and makes it look at that players home zone
   * @param playerNumber 0 for player 1, anything else for player 2
   */
  public final void SetCameraToPlayer(int playerNumber)
  {
    float camZ = this.ownerClient.getGameState().getTabledepth() * 0.85f;
    float lookZ = this.ownerClient.getGameState().getTabledepth() / 4f;
    
    if(playerNumber == 0)
    {
      camZ = -camZ;
      lookZ = -lookZ;
    }
    
    world.getCamera().setFOV(1f);
    world.getCamera().setPosition(0f, -this.ownerClient.getGameState().getTabledepth() / 4f, camZ);
    world.getCamera().lookAt(new SimpleVector(0.0f, 0.0f, lookZ));
  }

  /** 
   * centers the mouse in the view
   * (used by the mouse motion listener to keep determining relative mouse movement)
   */
  public void CenterMouse()
  {
    // center cursor in view
    int x = this.getLocationOnScreen().x + this.getWidth() / 2;
    int y = this.getLocationOnScreen().y + this.getHeight() / 2;
    robot.mouseMove(x, y);
  }
  
  /**
   * implementation of IPaddlePuckClientView.SignalUpdate
   * (will get called from the regular updates thread in the PaddlePuckClient)
   */
  @Override
  public synchronized void SignalUpdate()
  {
    this.paint(this.getGraphics());
  }
  
  /**
   * implementation of IPaddlePuckClientView.SignalServerDisconnected
   * (will get called from the PaddlePuckClient if the server disconnects)
   */
  @Override
  public synchronized void SignalServerDisconnected()
  {
    ((PaddlePuckClientGUI)this.tlFrame).Disconnect();
  }
  
  @Override
  public void paint(Graphics g)
  {
    if(skipPaint)
      return;
    
    skipPaint = true;
    
    int score0 = 0;
    int score1 = 0;
    int seconds = this.ownerClient.getSecondsLeft();
    
    synchronized(this.ownerClient.getGameState())
    {
      // update paddles from gamestate
      float[] pos = this.ownerClient.getGameState().getPaddlePosition(0);
      paddle[0].clearTranslation();
      paddle[0].translate(pos[0], pos[1], pos[2]);
      pos = this.ownerClient.getGameState().getPaddlePosition(1);
      paddle[1].clearTranslation();
      paddle[1].translate(pos[0], pos[1], pos[2]);
      
      // update puck from gamestate
      pos = this.ownerClient.getGameState().getPuckPosition();
      puck.clearTranslation();
      puck.translate(pos[0], pos[1], pos[2]);
      
      // set puck transparency based on puckState
      if(this.ownerClient.getGameState().getPuckState() != 0)
      {
        puck.setTransparency(0); 
        puckShadow.clearTranslation();
        puckShadow.translate(pos[0], 0.0f, pos[2]);
        puckShadow.setVisibility(true);
        puckShadow.setTransparency(0);
      }
      else
      {
        puck.setTransparency(-1);
        puckShadow.setVisibility(false);
      }
      
      // update scores from gamestate
      score0 = this.ownerClient.getGameState().getScore(0);
      score1 = this.ownerClient.getGameState().getScore(1);
    }
    
    // render world
    fb.clear(Color.BLACK); // clear framebuffer
    world.renderScene(fb); // process geometry
    world.draw(fb); // draw processed geometry to backbuffer
    
    fb.update(); // update internal frontbuffer from internal backbuffer
    
    // draw scores on internal frontbuffer
    Graphics fbg = fb.getGraphics();
    fbg.setFont(this.getFont());
    
    fbg.setColor(Color.GREEN); 
    fbg.drawString(String.format("Player ONE: %1$d", score0), 0, 12);
    
    fbg.setColor(Color.CYAN); 
    fbg.drawString(String.format("Player TWO: %1$d", score1), 0, 24);
    
    // draw match time or pre match countdown
    
    fbg.setColor(Color.WHITE);
    if(seconds < 0)
    {
      fbg.drawString(String.format("next match starts in %1$ds", -(seconds)), 0, 36);
    }
    else
    {
      fbg.drawString(String.format("%1$02d:%2$02d", seconds / 60, seconds%60), 0, 36);
    }
    
    fb.display(g); // render frontbuffer into graphics object of this Canvas
      
    skipPaint = false;
  }
  
  /**
   * handles the mouse clicked event
   * to capture/un-capture the mouse inside the view
   * @param e 
   */
  private void mouseClickedHandler(MouseEvent e)
  {
    if(e.getButton() == 1) // left button
    {
      setMouseCaptured(!mouseCaptured);
    }
  }
  
  /**
   * sets the state of the mouseCaptured field
   * and updates cursor accordingly
   * @param mCaptured new state for the mouseCaptured field
   */
  public synchronized void setMouseCaptured(boolean mCaptured)
  {
    mouseCaptured = mCaptured; // toggle mouse capture
    
    if(mouseCaptured)
    {
      // hide mouse cursor
      tlFrame.setCursor(noCursor);
      CenterMouse();
    }
    else
    {
      // show mouse cursor
      tlFrame.setCursor(Cursor.getDefaultCursor());
    }
  }
  
  /**
   * handles the mouse moved event
   * (if mouse is captured, mouse position is re-centered all the time to detect relative movement for paddle control)
   * @param e 
   */
  private void mouseMovedHandler(MouseEvent e)
  {
    if(mouseCaptured)
    {
      int cx = this.getLocationOnScreen().x + this.getWidth() / 2;
      int cy = this.getLocationOnScreen().y + this.getHeight() / 2;
      
      // compute relative mouse movement (and scale it down)
      float scale = 0.3f;
      float dx = (e.getXOnScreen() - cx) * scale;
      float dy = (e.getYOnScreen() - cy) * scale;
      
      if(dx != 0 || dy != 0)
      {
        // accumulate paddle displacement in client
        this.ownerClient.AccumulateDesiredPaddleDisplacement(dx, -dy);
        
        // this.logging.LogMessage(String.format("dx: %1$d dy: %2$d", dx, dy));
        CenterMouse();
      }
    }
  }
   
  /**
   * creates the center piece face for the table
   * @param textureName must be a name previously registered with the texture manager
   * @return 3D object meant as the center piece
   */
  private Object3D createTableCenterPiece(String textureName)
  {
    Object3D r = new Object3D(2);
    
    float twh = this.ownerClient.getGameState().getTablewidth() / 2f;
    float tdh = this.ownerClient.getGameState().getTabledepth() / 2f;
    float hzd = this.ownerClient.getGameState().getHomezonedepth();
    
    SimpleVector[] sv = new SimpleVector[4];
    sv[0] = new SimpleVector(-twh, 0f, tdh-hzd);
    sv[1] = new SimpleVector(twh, 0f, tdh-hzd);
    sv[2] = new SimpleVector(twh, 0f, -tdh+hzd);
    sv[3] = new SimpleVector(-twh, 0f, -tdh+hzd);
    
    r.addTriangle(sv[0], sv[2], sv[1]);
    r.addTriangle(sv[0], sv[3], sv[2]);
    
    r.setTexture(textureName);
    r.setEnvmapped(true);
    r.build();
    
    return r;
  }
  
  /**
   * creates the home zone face (for either side of the table)
   * @param whichOne must be set to 0 or 1 (otherwise null will be returned)
   * @param textureName must be a name previously registered with the texture manager
   * @return 3D object meant as a home zone
   */
  private Object3D createTableHomeZone(int whichOne, String textureName)
  {
    if(whichOne != 0 && whichOne != 1)
      return null;
    
    Object3D r = new Object3D(2);
    
    float twh = this.ownerClient.getGameState().getTablewidth() / 2f;
    float tdh = this.ownerClient.getGameState().getTabledepth() / 2f;
    float hzd = this.ownerClient.getGameState().getHomezonedepth();
    
    SimpleVector[] sv = new SimpleVector[4];
    if(whichOne == 0)
    {
      sv[0] = new SimpleVector(-twh, 0f, -tdh+hzd);
      sv[1] = new SimpleVector(twh, 0f, -tdh+hzd);
      sv[2] = new SimpleVector(twh, 0f, -tdh);
      sv[3] = new SimpleVector(-twh, 0f, -tdh);
    }
    else if(whichOne == 1)
    {
      sv[0] = new SimpleVector(-twh, 0f, +tdh);
      sv[1] = new SimpleVector(twh, 0f, +tdh);
      sv[2] = new SimpleVector(twh, 0f, +tdh-hzd);
      sv[3] = new SimpleVector(-twh, 0f, +tdh-hzd);
    }
    
    r.addTriangle(sv[0], sv[2], sv[1]);
    r.addTriangle(sv[0], sv[3], sv[2]);
    
    r.setTexture(textureName);
    r.setEnvmapped(true);
    r.build();
    
    return r;
  }
  
  /**
   * creates a paddle
   * @param textureName must be a name previously registered with the texture manager
   * @return 3D object meant as a paddle
   */
  private Object3D createPaddleObject(String textureName)
  {
    Object3D r = new Object3D(12);
    
    float pdh = this.ownerClient.getGameState().getPaddledepth() / 2f;
    float pwh = this.ownerClient.getGameState().getPaddlewidth() / 2f;
    float h = this.ownerClient.getGameState().getPaddleheight();
    
    SimpleVector[] sv = new SimpleVector[8];
    
    sv[0] = new SimpleVector(-pwh, -h, -pdh);
    sv[1] = new SimpleVector(pwh, -h, -pdh);
    sv[2] = new SimpleVector(pwh, 0f, -pdh);
    sv[3] = new SimpleVector(-pwh, 0f, -pdh);
    sv[4] = new SimpleVector(-pwh, -h, pdh);
    sv[5] = new SimpleVector(pwh, -h, pdh);
    sv[6] = new SimpleVector(pwh, 0f, pdh);
    sv[7] = new SimpleVector(-pwh, 0f, pdh);
   
    r.addTriangle(sv[0], sv[2], sv[1]); // frontface
    r.addTriangle(sv[0], sv[3], sv[2]);
    r.addTriangle(sv[5], sv[7], sv[4]); // backface
    r.addTriangle(sv[5], sv[6], sv[7]);
    r.addTriangle(sv[4], sv[3], sv[0]); // leftface
    r.addTriangle(sv[4], sv[7], sv[3]);
    r.addTriangle(sv[1], sv[6], sv[5]); // rightface
    r.addTriangle(sv[1], sv[2], sv[6]);
    r.addTriangle(sv[3], sv[6], sv[2]); // bottomface
    r.addTriangle(sv[3], sv[7], sv[6]);
    r.addTriangle(sv[4], sv[1], sv[5]); // topface
    r.addTriangle(sv[4], sv[0], sv[1]);
    
    r.setTexture(textureName);
    r.setEnvmapped(true);
    r.build();
    
    return r;
  }

  /**
   * creates an approximation of a round puck
   * @param createShadow set to true to create puck shadow instead of the puck itself
   * @param sideFaces the number of faces on the side of the approximation
   * @param textureName must be a name previously registered with the texture manager
   * @return 3D object meant as the puck
   */
  private Object3D createPuckObject(boolean createShadow, int sideFaces, String textureName)
  {
    int faces = (sideFaces >= 4) ? sideFaces : 4;
    int m = createShadow ? 1 : 4;
    Object3D ret = new Object3D(faces * m); // two triangles for each side face and another two for each side face (for the top/bottom triangle fans)
    
    float ph = this.ownerClient.getGameState().getPuckheight();
    
    if(createShadow)
      ph = 0.15f;
    
    float r = this.ownerClient.getGameState().getPuckradius();
    float alpha = 2f * (float)Math.PI / (float)faces;
    
    SimpleVector svBotCenter = new SimpleVector(0.0f, 0.0f, 0.0f);
    SimpleVector svTopCenter = new SimpleVector(0.0f, -ph, 0.0f);
    SimpleVector[] svTopCircle = new SimpleVector[faces];
    SimpleVector[] svBotCircle = new SimpleVector[faces];
    
    // create circle approximations
    for(int i=0; i<faces; i++)
    {
      svTopCircle[i] = new SimpleVector(r * Math.cos(i*alpha), -ph, r * Math.sin(i*alpha));
      
      if(!createShadow)
        svBotCircle[i] = new SimpleVector(svTopCircle[i].x, 0.0f, svTopCircle[i].z);
    }
    
    // create side/top/bottom faces
    for(int i=0; i<faces; i++)
    {
      int j = (i + 1)%faces;
      
      // create sideface
      if(!createShadow)
      {
        ret.addTriangle(svTopCircle[j], svTopCircle[i], svBotCircle[i]);
        ret.addTriangle(svTopCircle[j], svBotCircle[i], svBotCircle[j]);
      }
      
      // create top face
      ret.addTriangle(svTopCenter, svTopCircle[i], svTopCircle[j]);
      
      // create bottom face
      if(!createShadow)
        ret.addTriangle(svBotCenter, svBotCircle[j], svBotCircle[i]);
    }
    
    ret.setTexture(textureName);
    ret.setEnvmapped(true);
    ret.build();
    
    return ret;
  }

  /**
   * creates a table border
   * @param leftOrRight pass "l" or "r" (anything else will create the border centered)
   * @param textureName must be a name previously registered with the texture manager
   * @return 3D object meant as a table border
   */
  private Object3D createTableBorder(String leftOrRight, String textureName)
  {
    Object3D r = new Object3D(12);
    
    float bdh = this.ownerClient.getGameState().getTabledepth() / 2f;
    float bwh = this.ownerClient.getGameState().getTablewidth() / 8f;
    float twh = this.ownerClient.getGameState().getTablewidth() / 2f;
    float h = this.ownerClient.getGameState().getPaddleheight();
    
    SimpleVector[] sv = new SimpleVector[8];
    
    sv[0] = new SimpleVector(-bwh, -h, -bdh);
    sv[1] = new SimpleVector(bwh, -h, -bdh);
    sv[2] = new SimpleVector(bwh, 0f, -bdh);
    sv[3] = new SimpleVector(-bwh, 0f, -bdh);
    sv[4] = new SimpleVector(-bwh, -h, bdh);
    sv[5] = new SimpleVector(bwh, -h, bdh);
    sv[6] = new SimpleVector(bwh, 0f, bdh);
    sv[7] = new SimpleVector(-bwh, 0f, bdh);
   
    r.addTriangle(sv[0], sv[2], sv[1]); // frontface
    r.addTriangle(sv[0], sv[3], sv[2]);
    r.addTriangle(sv[5], sv[7], sv[4]); // backface
    r.addTriangle(sv[5], sv[6], sv[7]);
    r.addTriangle(sv[4], sv[3], sv[0]); // leftface
    r.addTriangle(sv[4], sv[7], sv[3]);
    r.addTriangle(sv[1], sv[6], sv[5]); // rightface
    r.addTriangle(sv[1], sv[2], sv[6]);
    r.addTriangle(sv[3], sv[6], sv[2]); // bottomface
    r.addTriangle(sv[3], sv[7], sv[6]);
    r.addTriangle(sv[4], sv[1], sv[5]); // topface
    r.addTriangle(sv[4], sv[0], sv[1]);
    
    r.setTexture(textureName);
    r.setEnvmapped(true);
    r.build();
    
    if(leftOrRight.compareTo("l") == 0)
      r.translate(new SimpleVector(-twh-bwh, 0f, 0f));
    else if(leftOrRight.compareTo("r") == 0)
      r.translate(new SimpleVector(twh+bwh, 0f, 0f));
    
    return r;
  }
}
