/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckprotocol;

import util.Logging;
import util.SimpleRunnable;
import paddlepuckGUI.IPaddlePuckClientView;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;

/**
 * The Paddle Puck Client.
 */
public final class PaddlePuckClient
{
  private Logging logging = null;
  
  public static volatile int Version = 100;
  private int serverPort = 42000;
  private Socket clientSocket = null;
  
  private final Integer serverLock = new Integer(0);
  private PaddlePuckConnectionClientSide server = null;
  private final PaddlePuckGameState gamestate = new PaddlePuckGameState();
  
  // for input evaluation / regularly sending input to server / regularly signalling to update view
  private IPaddlePuckClientView ippcw = null;
  private Thread sendInputSignalView = null;
  private final int updatesPerSecond = PaddlePuckGameState.CyclesPerSecond;
  
  private final Integer secondsLock = new Integer(0);
  private int secondsLeft = 0;
  
  private final Integer deltaLock = new Integer(0);
  private float pdx = 0f; // paddle desired delta x
  private float pdz = 0f; // paddle desired delta z
  
  /**
   * Constructs the PaddlePuckClient object (but does not connect to 
   * the server right away)
   * @param logging the logging object for logging messages
   * @throws Exception could be thrown during thread setup
   */
  public PaddlePuckClient(Logging logging) throws Exception
  {
    this.logging = logging;
      
    // initialize "regular input send and signal view" thread
    this.sendInputSignalView = SimpleRunnable.CreateThread(this, "SendInputAndSignalViewThread", "sendInputSignalView");
  }
  
  /**
   * sets the view (will only work while the regular updates thread is not running)
   * @param ippcv must reference the view which will be signaled regularly to update itself (setting this to null disables notification of updates)
   */
  public void setView(IPaddlePuckClientView ippcv)
  {
    if(!this.sendInputSignalView.isAlive())
      this.ippcw = ippcv;
  }
  
  public void StartRegularUpdatesThread()
  {
    this.sendInputSignalView.start();
  }
  
  public void StopRegularUpdatesThread()
  {
    this.sendInputSignalView.interrupt();
  }
  
  /**
   * Connects to the server and starts processing server messages
   * in the PaddlePuckConnectionClientSide object thread for that.
   * @param host describes which server to connect to
   * @param port describes the port on which the server listens/accepts incoming connections
   * @param localPort describes the port on which the client will send/receive server messages
   * @throws IOException as thrown in the PaddlePuckConnection constructor on getting streams from a socket connection
   * @throws Exception as thrown on problems during CONNECTION SETUP protocol
   */
  public void ConnectToServer(String host, int port, int localPort) throws IOException, Exception
  {
    this.serverPort = port;
    
    clientSocket = new Socket();
    clientSocket.setReuseAddress(true);
    //clientSocket.bind(new InetSocketAddress(localPort)); // removed because reuse of specific client port never worked and threw and exception on .connect
    //clientSocket.bind(null);
    clientSocket.connect(new InetSocketAddress(host, this.serverPort), 5000);
    
    server = new PaddlePuckConnectionClientSide(this, clientSocket, Version, logging); 
    
    this.logging.LogMessage(String.format("Connected as player %1$s", this.server.getPlayerNumber() == 0 ? "ONE" : "TWO"));
    
    server.StartProcessingServerMessages();
  }
  
  /**
   * Disconnects from the server and stop processing server messages
   * in the PaddlePuckConnectionClientSide object thread for that.
   */
  public void DisconnectFromServer()
  {  
    synchronized(serverLock)
    {
      if (server != null)
      {
        try
        {
          // send message to server to close connection on their side
          server.SendCloseConnection();

          server.Close();
          server.StopProcessingServerMessages(true);
        }
        catch(Exception e)
        {
          this.logging.LogException(e);
        }
        finally
        {
          server = null;
          this.logging.LogMessage("Disconnected.");
        }
      }   
    }
    
    // close client socket
    try
    {
      this.logging.LogMessage("closing client socket...");
      this.clientSocket.close();
    }
    catch(Exception e)
    {
      this.logging.LogException(e);
    }
  }
  
  /**
   * get the player number which was assigned from the server during CONNECTION SETUP
   * @return -1 if the the connection hasn't been set up or was disconnected
   */
  public int getPlayerNumber()
  {
    synchronized(serverLock)
    {
      if(this.server != null)
        return this.server.getPlayerNumber();
      else
        return -1;
    }
  }
  
  /**
   * returns true if this client is theoretically still connected to a server
   * @return false if there is no connection
   */
  public boolean isConnected()
  {
    return this.getPlayerNumber() != -1;
  }
  
  /**
   * get the game state
   */
  public PaddlePuckGameState getGameState()
  {
    return this.gamestate;
  }
  
  /**
   * get the last set secondsLeft value
   * (was either set by message from server or calculated locally
   *  in SendInputAndSignalViewThread)
   */
  public int getSecondsLeft()
  {
    int r = 0;
    synchronized(secondsLock)
    {
      r = this.secondsLeft;
    }
    return r;
  }
  
  /**
   * this thread regularly sends update signals to the client view (if there was only supplied in the constructor)
   * and it also regularly sends the collected input data to the server (if it's connected)
   */
  public void SendInputAndSignalViewThread()
  {
    while(true)
    {
      try
      {
        
        if(this.isConnected()) // if connected
        {
          try
          {
            float spdx, spdz;
            synchronized(deltaLock)
            {
              spdx = pdx;
              spdz = pdz;
              this.resetDesiredPaddleDisplacement();
            }
              
            // send desire to move paddle to server
            if(spdx != 0f || spdz != 0f)
            {
              synchronized(this.serverLock)
              {
                this.server.SendDesiredPaddleDisplacement(spdx, spdz);
              }
            }
          }
          catch(IOException ioe)
          {
            this.logging.LogMessage("failed to send paddle input to server!");
            this.logging.LogException(ioe);
          }
        }
        else // if not connected
        {
          synchronized(deltaLock)
          {
            // (just move paddle and update gamestate locally)
            this.gamestate.setDesiredPaddleDisplacement(0, pdx, pdz);
            this.gamestate.Update(true);
            
            // reset displacement
            this.resetDesiredPaddleDisplacement();
          }
          
          synchronized(secondsLock)
          {             
            secondsLeft = this.gamestate.getSecondsLeft();
          }
        }
        
        // send update signal to client view (if it exists)
        if(this.ippcw != null)
          this.ippcw.SignalUpdate();
        
        Thread.sleep(1000 / updatesPerSecond);
      }
      catch(InterruptedException ie)
      {
        this.logging.LogException(ie);
        break; // end thread
      }
    }
  }
  
  /**
   * adds the given values to the internal desired paddle deltas
   * (which are periodically sent and cleared in SendInputAndSignalViewThread)
   * @param adx
   * @param adz 
   */
  public void AccumulateDesiredPaddleDisplacement(float adx, float adz)
  {
    synchronized(deltaLock)
    {
      this.pdx += adx;
      this.pdz += adz;
    }
  }
  
  /**
   * resets the desired paddle displacement to none
   */
  private void resetDesiredPaddleDisplacement()
  {
    synchronized(deltaLock)
    {
      this.pdx = 0.0f;
      this.pdz = 0.0f;
    }
  }
  
  /**
   * processes the CloseConnection message received in a PaddlePuckConnectionClientSide processIncomingMessages thread
   */
  public void HandleServerDisconnectedByMessage()
  {
    synchronized(serverLock)
    {
      try
      {
        server.Close();
        server.StopProcessingServerMessages(false);
        this.logging.LogMessage("Server disconnected.");
      }
      catch(Exception e)
      {
        this.logging.LogException(e);
      }
      finally
      {
        server = null;
        
        // send disconnected signal to client view
        if(this.ippcw != null)
          this.ippcw.SignalServerDisconnected();
      }
    }
  }
  
  /**
   * processes the SocketTimeoutException in a PaddlePuckConnectionClientSide processIncomingMessages thread
   */
  public void HandleServerDisconnectedByTimeout()
  {
    this.logging.LogMessage("Disconnecting from server due to SocketTimeoutException...");
    this.HandleServerDisconnectedByMessage();
  }

  /**
   * processes the EOFException in a PaddlePuckConnectionClientSide processIncomignMessages thread
   */
  public void HandleServerDisconnectedByEndOfStream()
  {
    this.logging.LogMessage("Disconnecting from server due to EOFException on incoming data stream...");
    this.HandleServerDisconnectedByMessage();
  }
  
  /**
   * processes the IOException in a PaddlePuckConnectionClientSide processIncomignMessages thread
   */
  public void HandleServerDisconnectedByIOException()
  {
    this.logging.LogMessage("Disconnecting from server due to IOException on incoming data stream...");
    this.HandleServerDisconnectedByMessage();
  }
  
  /**
   * processes the positions of GameStatePositionsUpdate message
   * @param p { paddle0 x, paddle0 y, paddle0 z, paddle1 x, paddle1 y, paddle1 z, puck x, puck y, puck z }
   */
  public void HandleGameStatePositionsUpdate(float[] p)
  {
    synchronized(this.gamestate)
    {
      this.gamestate.setPaddlePosition(0, p[0], p[1], p[2]);   
      this.gamestate.setPaddlePosition(1, p[3], p[4], p[5]);
      this.gamestate.setPuckPosition(p[6], p[7], p[8]);
    }
  }

  /**
   * processes the puck state of GameStatePuckStateUpdate message
   * @param ps new puck state
   */
  public void HandleGameStatePuckStateUpdate(int ps)
  {
    synchronized(this.gamestate)
    {
      this.gamestate.setPuckState(ps);
    }
  }
  
  /**
   * processes the scores of GameStateScoresUpdate message
   * @param s new scores
   */
  public void HandleGameStateScoresUpdate(int[] s)
  {
    synchronized(this.gamestate)
    {
      this.gamestate.setScore(0, s[0]);
      this.gamestate.setScore(1, s[1]);
    }
  }
  
  /**
   * processes the GameStateMatchSecondsLeft message data
   * @param sleft the number of seconds left
   */
  public void HandleGameStateMatchSecondsLeft(int sleft)
  {
    synchronized(this.secondsLock)
    {
      this.secondsLeft = sleft;
    }
  }
}
