/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckprotocol;

import util.Logging;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;

/**
 * This class holds information about a single Client/Server connection end for
 * Paddle Puck.
 */
public class PaddlePuckConnection
{
  protected final Integer fromLock = new Integer(0);
  protected final Integer toLock = new Integer(0);
  
  protected Logging logging = null;
  
  protected Socket socket = null;
  protected DataOutputStream toStream = null;
  protected DataInputStream fromStream = null;
  protected PaddlePuckConnectionState connectionState = PaddlePuckConnectionState.NotConnected;
  
  protected int playerNumber = -1;
  
  /**
   * Constructs a PaddlePuckConnection object and performs the connection initialization protocol.
   * @param socket must be a previously initialized socket
   * @param logging the object used for logging messages
   * @throws IOException exception from creating output and input streams on the socket
   */
  public PaddlePuckConnection(Socket socket, Logging logging) throws IOException
  {
    this.logging = logging;
    
    this.socket = socket;
    connectionState = PaddlePuckConnectionState.NotConnected;
    
    if(this.socket != null)
    {
      toStream = new DataOutputStream(this.socket.getOutputStream());
      fromStream = new DataInputStream(this.socket.getInputStream());
      
      connectionState = PaddlePuckConnectionState.Setup;
    }
  }
  
  /**
   * Closes the input and output streams of the socket and then closes the socket.
   * @throws IOException exception from closing input and output streams of the socket
   */
  public final void Close() throws IOException
  {
    if( this.socket != null )
    {
      if (this.toStream != null)
        this.toStream.close();
      this.toStream = null;
      
      if (this.fromStream != null)
        this.fromStream.close();
      this.fromStream = null;
      
      this.socket.close();
      this.socket = null;
      
      this.connectionState = PaddlePuckConnectionState.NotConnected;
    }
  }
    
  public final Socket getSocket()
  {
    return this.socket;
  }
  
  public final DataOutputStream getToStream()
  {
    return this.toStream;
  }
  
  public final DataInputStream getFromStream()
  {
    return this.fromStream;
  }
  
  public final PaddlePuckConnectionState getConnectionState()
  {
    return this.connectionState;
  }
  
  public final int getPlayerNumber()
  {
    return this.playerNumber;
  }
  
  /**
   * Sends the CloseConnection message to the other end of the connection.
   * @throws IOException 
   */
  public final void SendCloseConnection() throws IOException
  {
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.CloseConnection.ordinal());
      toStream.flush();
    }
  }
 
  /**
   * Sends the HeartBeat message to the other end of the connection.
   * @throws IOException 
   */
  public final void SendHeartbeat() throws IOException
  {
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.Heartbeat.ordinal());
      toStream.flush();
    }
  }
}
