/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckprotocol;

import util.Logging;
import util.SimpleRunnable;
import java.io.EOFException;
import java.io.IOException;
import java.net.Socket;
import java.net.SocketTimeoutException;

/**
 * This class handles the client side of a PaddlePuckConnection.
 * It runs a thread which receives and processes messages coming from the server side
 * and provides methods for sending messages to the server side.
 */
public class PaddlePuckConnectionClientSide extends PaddlePuckConnection
{
  protected volatile boolean processingIncomingMessages = false;
  protected Thread processIncomingMessages = null;
  
  protected PaddlePuckClient ownerClient = null;
  
  /**
   * Constructs the client side connection state, initializes (but does not start)
   * the incoming messages processing thread and performs the CONNECTION SETUP protocol
   * @param ownerClient must be set to the PaddlePuckClient which creates this object (will be used to process received server messages) 
   * @param socket must be supplied as a new Socket created to connect to the desired server
   * @param version will be used in CONNECTION SETUP for the version check between server and client
   * @param logging the logging object to use for logging
   * @throws IOException exception as thrown by parent class constructor while getting the connection streams
   * @throws Exception will be thrown on CONNECTION SETUP problems
   */
  public PaddlePuckConnectionClientSide(PaddlePuckClient ownerClient, Socket socket, int version, Logging logging) throws IOException, Exception       
  {
    super(socket, logging);
   
    // store ownerClient (will handle messages received by the processIncomingMessages thread)
    this.ownerClient = ownerClient;
    
    // thread setup
    processIncomingMessages = SimpleRunnable.CreateThread(this, "RunProcessIncomingServerMessagesThread" , "processIncomingServerMessages");
    processingIncomingMessages = false;
    
    this.socket.setSoTimeout(5000); // set timeout to 5 seconds
    
    // CONNECTION SETUP (CLIENT SIDE)
    // send client identification string
    toStream.writeUTF("PaddlePuckClient");
    toStream.flush();
    
    // receive server identification string
    String serverId = fromStream.readUTF();
    if (serverId.compareTo("PaddlePuckServer") != 0)
    {
      this.Close();
      throw new Exception("Outgoing connection was accepted by something other than a PaddlePuckServer.");
    }
 
    // send client version
    toStream.writeInt(version);
    toStream.flush();

    // receive server version
    int serverVersion = fromStream.readInt();

    // close connection on version mismatch
    if(serverVersion != version)
    {
      this.Close();
      throw new Exception(String.format("Version mismatch! client(%1$d) / server(%2$d)", version, serverVersion));
    }

    // receive player number
    this.playerNumber = fromStream.readInt();
    
    if (this.playerNumber == -1)
    {
      this.Close();
      throw new Exception("Connection denied: Too many players.");
    }
    
    connectionState = PaddlePuckConnectionState.Connected;
    // END OF CONNECTION SETUP
  }

  /**
   * starts the processIncomingMessages thread
   */
  public void StartProcessingServerMessages()
  {
    processIncomingMessages.start();
  }
  
  /**
   * stops the processIncomingMessages thread
   * @param join set this to false if the processing thread itself invokes this method
   * @throws InterruptedException exception as thrown while joining the thread
   */
  public void StopProcessingServerMessages(boolean join) throws InterruptedException
  {
    processingIncomingMessages = false;
    
    if(join)
    {
      this.logging.LogMessage(String.format("joining Thread:%1$s...", processIncomingMessages.getName()));
      processIncomingMessages.join();
    }
  }
  
  /**
   * processIncomingMessages thread
   * this thread waits for and processes messages received from the server
   */
  public void RunProcessIncomingServerMessagesThread()
  {
    int ppmType = 0;
    PaddlePuckMessage ppm = PaddlePuckMessage.Undefined;
    PaddlePuckMessage allppms[] = PaddlePuckMessage.values();
    
    this.processingIncomingMessages = true;
    
    while(processingIncomingMessages) // PROCESSING MESSAGES RECEIVED FROM SERVER
    {
      Thread.yield();
      
      try
      {
        synchronized(fromLock)
        {
          ppmType = fromStream.readInt(); // read message head (type)
          ppm = allppms[ppmType];

          // evaluate message head, read message data, handle message
          switch(ppm)
          {
            case Undefined:
            break;

            case Heartbeat:
              logging.LogMessage("received heartbeat from server");
              this.SendHeartbeat(); // client responds to heartbeat, so server knows it's still running
            break;

            case CloseConnection:
              ownerClient.HandleServerDisconnectedByMessage();
            break;
              
            case GameStatePositionsUpdate:
              float[] p = this.readGameStatePositionsUpdate();
              ownerClient.HandleGameStatePositionsUpdate(p);
              
            break;
              
            case GameStatePuckStateUpdate:
              int ps = this.readGameStatePuckStateUpdate();
              ownerClient.HandleGameStatePuckStateUpdate(ps);
            break;
              
            case GameStateScoresUpdate:
              int s[] = this.readGameStateScoresUpdate();
              ownerClient.HandleGameStateScoresUpdate(s);
            break;
              
            case GameStateMatchSecondsLeft:
              int sleft = this.readGameStateMatchSecondsLeft();
              ownerClient.HandleGameStateMatchSecondsLeft(sleft);
            break;
              
            default:
            break;
          }
        }
      }
      catch(EOFException eofe)
      {
        logging.LogException(eofe);
        if(processingIncomingMessages)
          ownerClient.HandleServerDisconnectedByEndOfStream();
      }
      catch(SocketTimeoutException ste)
      {
        logging.LogException(ste);
        if(processingIncomingMessages)
          ownerClient.HandleServerDisconnectedByTimeout();
      }
      catch(IOException ioe)
      {
        logging.LogException(ioe);
        if(processingIncomingMessages)
          ownerClient.HandleServerDisconnectedByIOException();
      }
    }
  }
  
  /**
   * Sends the DesiredPaddleDisplacement message to the server side
   * @param dx delta x by which client wants to move paddle
   * @param dz delta z by which client wants to move paddle
   * @throws IOException 
   */
  public final void SendDesiredPaddleDisplacement(float dx, float dz) throws IOException
  {
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.DesiredPaddleDisplacement.ordinal());
      toStream.writeFloat(dx);
      toStream.writeFloat(dz);
      toStream.flush();
    }
  }

  /**
   * reads the positions data of the GameStatePositionsUpdate message and returns it
   * @return { paddle0 x, paddle0 y, paddle0 z, paddle1 x, paddle1 y, paddle1 z, puck x, puck y, puck z }
   * @throws IOException could be thrown by stream operation failure
   */
  private float[] readGameStatePositionsUpdate() throws IOException
  {
    float[] p = new float[9];
    synchronized(fromLock)
    {
      for(int i=0; i<9; i++)
        p[i] = this.fromStream.readFloat();
    }
    return p;
  }
  
  /**
   * reads the puck state of the GameStatePuckStateUpdate message
   * @return the new puck state
   * @throws IOException could be thrown by stream operation failure
   */
  private int readGameStatePuckStateUpdate() throws IOException
  {
    int r = 0;
    synchronized(fromLock)
    {
      r = this.fromStream.readInt();
    }
    return r;
  }
  
  /**
   * reads the scores data of the GameStateScoresUpdate message
   * @return the new scores for both players
   * @throws IOException could be thrown by stream operation failure
   */
  private int[] readGameStateScoresUpdate() throws IOException
  {
    int[] s = new int[2];
    synchronized(fromLock)
    {
      s[0] = this.fromStream.readInt();
      s[1] = this.fromStream.readInt();
    }
    return s;
  }
  
  /**
   * reads the secondsLeft data of the GameStateMatchSecondsLeft message
   * @return the new secondsLeft value (negative for pre match phase, positive while match lasts)
   * @throws IOException could be thrown by stream operation failure
   */
  private int readGameStateMatchSecondsLeft() throws IOException
  {
    int sleft = 0;
    synchronized(fromLock)
    {
      sleft = this.fromStream.readInt();
    }
    return sleft;
  }
  
}
