/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckprotocol;

import util.Logging;
import util.SimpleRunnable;
import java.io.EOFException;
import java.io.IOException;
import java.net.Socket;
import java.net.SocketTimeoutException;

/**
 * This class handles the server side of a PaddlePuckConnection.
 * It runs a thread which receives and processes messages coming from the client side
 * and provides methods for sending messages to the client side.
 */
public class PaddlePuckConnectionServerSide extends PaddlePuckConnection
{
  protected volatile boolean processingIncomingMessages = false;
  protected Thread processIncomingMessages = null;

  protected PaddlePuckServer ownerServer = null;
  
  /**
   * Constructs the server side connection state, initializes (but does not start)
   * the incoming messages processing thread and performs the CONNECTION SETUP protocol
   * @param ownerServer must be set to the PaddlePuckServer which creates this object (will be used to process received client messages)
   * @param socket must be supplied from a ServerSocket.accept call
   * @param version will be used in CONNECTION SETUP for the version check between server and client
   * @param playerNumber will be send so client knows which player it is ((-1) means server is full)
   * @param logging the logging object to use for logging
   * @throws IOException exception as thrown by parent class constructor while getting the connection streams
   * @throws Exception will be thrown on CONNECTION SETUP problems
   */
  public PaddlePuckConnectionServerSide(PaddlePuckServer ownerServer, Socket socket, int version, int playerNumber, Logging logging) throws IOException, Exception       
  {
    super(socket, logging);
    
    // store ownerServer (will handle messages received by the processIncomingMessages thread)
    this.ownerServer = ownerServer;
    
    // thread setup
    processIncomingMessages = SimpleRunnable.CreateThread(this, "RunProcessIncomingMessagesThread", String.format("processIncomingMessages%1$d", playerNumber));
    processingIncomingMessages = false;
    
    this.socket.setSoTimeout(5000); // set timeout to 5 seconds
    
    // CONNECTION SETUP (SERVER SIDE)
    // recieve client identification string
    String clientId = fromStream.readUTF();
    if(clientId.compareTo("PaddlePuckClient") != 0)
    {
      this.Close();
      throw new Exception("Incoming connection was not from a PaddlePuckClient.");
    }
    
    // send server identification string
    toStream.writeUTF("PaddlePuckServer");
    toStream.flush();
   
    // receive client version
    int clientVersion = fromStream.readInt();

    // send server version
    toStream.writeInt(version);
    toStream.flush();

    // close connection on version mismatch
    if(clientVersion != version)
    {
      this.Close();
      throw new Exception(String.format("Version mismatch! client(%1$d) / server(%2$d)", clientVersion, version));
    }

    // send player number
    this.playerNumber = playerNumber;
    toStream.writeInt(this.playerNumber);
    toStream.flush();
    
    // close connection if server is full
    if(this.playerNumber == -1)
    {
      this.Close();
      throw new Exception("Denied an incoming connection: Too many players.");
    }
    
    connectionState = PaddlePuckConnectionState.Connected;
    // END OF CONNECTION SETUP
  }

  /**
   * starts the processIncomingMessages thread
   */
  public void StartProcessingClientMessages()
  {
    processIncomingMessages.start();
  }
  
  /**
   * stops the processIncomingMessages thread
   * @param join set this to false if the processing thread itself invokes this method
   * @throws InterruptedException exception as thrown while joining the thread
   */
  public void StopProcessingClientMessages(boolean join) throws InterruptedException
  {
    processingIncomingMessages = false;
    
    if(join)
    {
      this.logging.LogMessage(String.format("joining Thread:%1$s...", processIncomingMessages.getName()));
      processIncomingMessages.join();
    }
  }
  
  /**
   * processIncomingMessages thread
   * this thread waits for and processes messages received from the client
   */
  public void RunProcessIncomingMessagesThread()
  {
    int ppmType = 0;
    PaddlePuckMessage ppm = PaddlePuckMessage.Undefined;
    PaddlePuckMessage allppms[] = PaddlePuckMessage.values();
    
    this.processingIncomingMessages = true;
    
    while(processingIncomingMessages) // PROCESSING MESSAGES RECEIVED FROM CLIENT
    {
      Thread.yield();
      
      try
      {
        synchronized(fromLock)
        {
          ppmType = fromStream.readInt(); // read message head (type)
          ppm = allppms[ppmType];

          // evaluate message head, read message data, handle message
          switch(ppm)
          {
            case Undefined:
            break;

            case Heartbeat:
              logging.LogMessage(String.format("received heartbeat from client[%1$d]", this.playerNumber));
              // server does not respond to heartbeat (PaddlePuckServer runs a thread which initiates/sends heartbeats to clients which then respond with a heartbeat to the server)
            break;

            case CloseConnection:
              ownerServer.HandleClientDisconnectedByMessage(this.playerNumber);
            break;
              
            case DesiredPaddleDisplacement:
              float[] d = this.readDesiredPaddleDisplacement();
              ownerServer.HandleDesiredPaddleDisplacementMessage(this.playerNumber, d[0], d[1]);
            break;

            default:
            break;
          }
        }
      }
      catch(EOFException eofe)
      {
        logging.LogException(eofe);
        if(processingIncomingMessages)
          ownerServer.HandleClientDisconnectedByEndOfStream(this.playerNumber);
      }
      catch(SocketTimeoutException ste)
      {
        logging.LogException(ste);
        if(processingIncomingMessages)
          ownerServer.HandleClientDisconnectedByTimeout(this.playerNumber);
      }
      catch(IOException ioe)
      {
        logging.LogException(ioe);
        if(processingIncomingMessages)
          ownerServer.HandleClientDisconnectedByIOError(this.playerNumber);
      }
    }
  }
  
  /**
   * reads the data of the DesiredPaddleDisplacement message and returns it
   * @return { delta x, delta z }
   * @throws IOException could be thrown by stream operation failure
   */
  private float[] readDesiredPaddleDisplacement() throws IOException
  {
    float[] r = new float[2];
    synchronized(fromLock)
    {
      r[0] = this.fromStream.readFloat();
      r[1] = this.fromStream.readFloat();
    }
    return r;
  }
  
  /**
   * Sends the GameStatePositionsUpdate message
   * @param gs game state from which to get the positions to send
   * @throws IOException could be thrown by stream operation failure
   */
  public void SendGameStatePositionsUpdate(PaddlePuckGameState gs) throws IOException
  {  
    float[] p = new float[9];
    
    // grab positions from gamestate
    synchronized(gs)
    {
      float[] paddle0pos = gs.getPaddlePosition(0);
      float[] paddle1pos = gs.getPaddlePosition(1);
      float[] puckpos = gs.getPuckPosition();
      for(int i=0; i<3; i++)
      {
        p[i] = paddle0pos[i];
        p[i+3] = paddle1pos[i];
        p[i+6] = puckpos[i];
      }
    }
    
    // send positions
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.GameStatePositionsUpdate.ordinal());
      for(int i=0; i<9; i++)
        toStream.writeFloat(p[i]);
     
      toStream.flush();
    }
  }
  
  /**
   * Sends the GameStatePuckStateUpdate message
   * @param gs game state from which to get the puck state to send
   * @throws IOException could be thrown by stream operation failure
   */
  public void SendGameStatePuckStateUpdate(PaddlePuckGameState gs) throws IOException
  {
    int ps = 0;
    
    // grab puck state from game state
    synchronized(gs)
    {
      ps = gs.getPuckState();
    }
    
    // send puck state
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.GameStatePuckStateUpdate.ordinal());
      toStream.writeInt(ps);
      toStream.flush();
    }
  }
  
  /**
   * Sends the GameStateScoresUpdate message
   * @param gs game state from which to get the scores to send
   * @throws IOException could be thrown by stream operation failure
   */
  public void SendGameStateScoresUpdate(PaddlePuckGameState gs) throws IOException
  {
    int[] score = new int[2];
    
    // grab scores from game state
    synchronized(gs)
    {
      score[0] = gs.getScore(0);
      score[1] = gs.getScore(1);
    }
    
    // send scores
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.GameStateScoresUpdate.ordinal());
      toStream.writeInt(score[0]);
      toStream.writeInt(score[1]);
      toStream.flush();
    }
  }
  
  /**
   * Sends the GameStateMatchSecondsLeft message
   * @param gs game state from which to get the seconds to send
   * @throws IOException could be thrown by stream operation failure
   */
  public void SendGameStateMatchSecondsLeft(PaddlePuckGameState gs) throws IOException
  {
    int sleft = 0;
    
    // grab value from game state
    synchronized(gs)
    {
      sleft = gs.getSecondsLeft();
    }
    
    // send value
    synchronized(toLock)
    {
      toStream.writeInt(PaddlePuckMessage.GameStateMatchSecondsLeft.ordinal());
      toStream.writeInt(sleft);
      toStream.flush();
    }
  }
}
