/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckprotocol;

import util.MyMath.*;

/**
 * This class holds all the book-keeping for a single PaddlePuck match.
 * It also contains the game update logic.
 */
public final class PaddlePuckGameState
{
  public static final int CyclesPerSecond = 50;
  private int gamecycle = 0;
  
  // size definitions (in cm)
  private float paddlewidth = 20.0f;
  private float paddledepth =  5.0f;
  private float paddleheight = 5.0f;
  
  private float puckheight = 2.54f; // DEL icehockey puck height (rule 250 puck)
  private float puckradius = 3.81f; // DEL icehockey puck radius (rule 250 puck)
  private float puckServeSpeedZ = 1.0f; // the z speed for the puck moving from the center towards the player who is next to serve
  private float puckRiseHeight = -2f*paddleheight; // rise height for the puck after anyone scores
  private float puckAscDescSpeed = 0.25f; // ascend/descend speed for puck after anyone scores
  
  private float tablewidth = 80.0f;
  private float tabledepth = 240.0f;
  private float homezonedepth = 30.0f;
  
  // paddle and puck position extreme values
  private float paddleXmin = -tablewidth/2f + paddlewidth/2f;
  private float paddleXmax = tablewidth/2f - paddlewidth/2f;
  private float[] paddleZmin = new float[] { (-tabledepth/2f + paddledepth/2f), (tabledepth/2f - homezonedepth + paddledepth/2f) }; // for paddle 0, paddle 1
  private float[] paddleZmax = new float[] { (-tabledepth/2f + homezonedepth - paddledepth/2f), (tabledepth/2f - paddledepth/2f) }; // for paddle 0, paddle 1
  
  // positions (in cm, in three dimensions)
  private float[][] paddleposition = new float[2][3]; // two paddle positions (client0,client1)(x,y,z)
  private Point2D[] prevPaddleposition = new Point2D[2]; // two previous game cycle paddle positions
  private float[] puckposition = new float[3]; // puck position (x,y,z)
  
  // puck movement vector
  private float[] puckspeed = new float[3];
  
  // desired paddle displacements (will be evaluated during update)
  private float[] pdx = new float[2];
  private float[] pdz = new float[2];
  
  // scores
  private int[] score = new int[2]; // two score counters (client0, client1)
  
  // match execution variables
  private final int totalMatchCycles = CyclesPerSecond * 60 * 5; // five minutes for n updates per second
  private final int preMatchCycles = CyclesPerSecond* 10; // ten second pre-match phase
  private int whoServesNext = 0; // player who is next to serve the puck
  private boolean puckCollision = true; // set to false to make puck un-collidable with anything
  private int puckState = 0; // 0 puck is in-game, 1 puck is ascending, 2 puck is moving towards center, 3 puck is descending
  
  
  /**
   * constructs a new game state object
   */
  public PaddlePuckGameState()
  {
    Reset();
  }

  /**
   * resets the whole game state
   */
  public void Reset()
  {
    // set paddle positions to defaults
    setPaddlePosition(0, 0.0f, 0.0f, 0.0f - 0.5f * tabledepth + 0.5f * homezonedepth); // centered in homezone 0
    setPaddlePosition(1, 0.0f, 0.0f, 0.0f + 0.5f * tabledepth - 0.5f * homezonedepth); // centered in homezone 1
    
    prevPaddleposition[0] = Point2D.from3DmapXZtoXY(getPaddlePosition(0));
    prevPaddleposition[1] = Point2D.from3DmapXZtoXY(getPaddlePosition(1));
    
    // set paddle displacements to defaults
    setDesiredPaddleDisplacement(0, 0.0f, 0.0f);
    setDesiredPaddleDisplacement(1, 0.0f, 0.0f);
    
    // reset the match
    this.resetMatch();
  }
  
  /**
   * resets only the match specific stuff but not the whole gamestate
   */
  private void resetMatch()
  {
    setPuckPosition(0.0f, puckRiseHeight, 0.0f); // center puck on table at drop height
    
    // set puck speed to serve speed
    setPuckSpeed(0.0f, 0.0f, -puckServeSpeedZ); // move towards player 0
    whoServesNext = 0; // player 0 will serve first
    puckCollision = false;
    puckState = 3;
    
    // reset scores
    score[0] = 0;
    score[1] = 0;
    
    // reset gamecycle to pre match phase
    gamecycle = -preMatchCycles;
  }
  
  /**
   * calculates and sanitizes the targetPosition for the given paddle
   * @param paddleNumber 0 or 1
   * @param xzPos the current x,z position for the given paddle
   * @param xzSpeed the desired x,z displacement for the given paddle
   * @return { x, z } of the sanitized target position
   */
  private Point2D getSanitizedPaddleTargetXZPosition(int paddleNumber, Point2D xzPos, Point2D xzSpeed)
  {
    Point2D stp = Cartesian2D.vectorSum(xzPos, xzSpeed);
    
    if(stp.x < paddleXmin)
      stp.x = paddleXmin;
    if(stp.x > paddleXmax)
      stp.x = paddleXmax;
    
    if(stp.y < paddleZmin[paddleNumber])
      stp.y = paddleZmin[paddleNumber];
    if(stp.y > paddleZmax[paddleNumber])
      stp.y = paddleZmax[paddleNumber];
    
    return stp;
  }
  
  /**
   * calculates and sanitizes the targetPosition for the puck
   * @param xzPos the current x,z position for the puck
   * @param xzSpeed the desired x,z displacement for the puck
   * @return { x, z } of the sanitized target position
   */
  private Point2D getSanitizedPuckTargetXZPosition(Point2D xzPos, Point2D xzSpeed)
  {
    Point2D stp = Cartesian2D.vectorSum(xzPos, xzSpeed);
    
    if(stp.x - puckradius < -tablewidth/2)
      stp.x = -tablewidth/2 + puckradius;
    if(stp.x + puckradius > tablewidth/2)
      stp.x = tablewidth/2 - puckradius;
    
    return stp;
  }
  
  /**
   * tests whether the puck would collide with any border if it was currently at the given target position
   * @param targetPuckXZ the target position for the puck
   * @return -1 at no collision, 0 border behind player 0, 1 border right of player 0, 2 border behind player 1, 3 border right of player 0
   */
  public int puckBorderCollision(Point2D targetPuckXZ)
  {
    if(!puckCollision)
      return -1;
    
    if (targetPuckXZ.y - puckradius <= -tabledepth/2)
      return 0;
    
    if (targetPuckXZ.x + puckradius >= tablewidth/2)
      return 1;
    
    if (targetPuckXZ.y + puckradius >= tabledepth/2)
      return 2;
    
    if (targetPuckXZ.x - puckradius <= -tablewidth/2)
      return 3;
    
    
    return -1;
  }
  
  /**
   * tests whether the puck and a paddle would collide if they were currently at the given target positions
   * 
   * 
   *    6-----------5------------4 
   *    |                        |
   *    7                        3
   *    |                        | 
   *    0-----------1------------2
   *   
   * @param targetPaddleXZ target position for the paddle
   * @param targetPuckXZ target position for the puck
   * @return -1 no collision, 0 to 7 if there is a collision (describing the vertex or edge of the paddle where the puck would collide with it)
   */
  public int puckPaddleCollision(Point2D targetPaddleXZ, Point2D targetPuckXZ)
  {
    if(!puckCollision)
      return -1;
    
    Point2D[] paddleVertex = new Point2D[4];
    paddleVertex[0] = Cartesian2D.vectorSum(targetPaddleXZ, new Point2D(-paddlewidth/2, -paddledepth/2));
    paddleVertex[1] = Cartesian2D.vectorSum(targetPaddleXZ, new Point2D(paddlewidth/2, -paddledepth/2));
    paddleVertex[2] = Cartesian2D.vectorSum(targetPaddleXZ, new Point2D(paddlewidth/2, paddledepth/2));
    paddleVertex[3] = Cartesian2D.vectorSum(targetPaddleXZ, new Point2D(-paddlewidth/2, paddledepth/2));
    
    // check if any paddle vertex would be inside puck circle
    for(int i=0; i<4; i++)
      if( Cartesian2D.vector(paddleVertex[i], targetPuckXZ).length() <= puckradius)
        return i*2;
  
    Point2D[] puckVertex = new Point2D[4];
    puckVertex[0] = new Point2D(targetPuckXZ.x, targetPuckXZ.y + puckradius);
    puckVertex[1] = new Point2D(targetPuckXZ.x - puckradius, targetPuckXZ.y);
    puckVertex[2] = new Point2D(targetPuckXZ.x, targetPuckXZ.y - puckradius);
    puckVertex[3] = new Point2D(targetPuckXZ.x + puckradius, targetPuckXZ.y);
    
    // check if any puck vertex would be inside the paddle rectangle
    for(int i=0; i<4; i++)
    {
      if(puckVertex[i].x >= paddleVertex[0].x &&
         puckVertex[i].x <= paddleVertex[1].x &&
         puckVertex[i].y >= paddleVertex[0].y &&
         puckVertex[i].y <= paddleVertex[2].y)
      return i*2 + 1;
    }
    
    return -1;
  }
  
  public void Update(boolean CPUmovePaddle1)
  {
    gamecycle++;
    
    if(gamecycle == this.totalMatchCycles) // match ends?
    {
      gamecycle = -preMatchCycles; // restart after another pre match phase
    }
    
    if(gamecycle < 0) // still or back in pre match phase?
    {
      setPuckPosition(0.0f, puckRiseHeight, 0.0f); // center puck on table at drop height
      puckCollision = false;
      puckState = 3;
    }
    else if(gamecycle == 0) // match starts?
    {
      this.resetMatch();
      gamecycle = 0;
    }
    
    // for testing move opposite paddle 1 in front of puck (when not connected)
    if(CPUmovePaddle1)
    {
      Point2D p1ToPuck = Cartesian2D.vector(Point2D.from3DmapXZtoXY(paddleposition[1]), Point2D.from3DmapXZtoXY(puckposition));
      pdx[1] = p1ToPuck.x;
    }
    
    // match execution handling
    if(puckState != 0)
    {
      setPuckSpeed(0f,0f,0f); // null puck speed so it's not affected by the regular puck movement methods
      
      if(puckState == 1) // puck ascending (after someone scored)
      {
        puckposition[1] -= puckAscDescSpeed;
        if(puckposition[1] <= puckRiseHeight) // puck near the max raise height?
        {
          puckposition[1] = puckRiseHeight;
          puckState = 2;
        }
      }
      else if(puckState == 2) // puck moving towards center (after someone scored, and puck ascended)
      {
        Point2D puPosNow = Point2D.from3DmapXZtoXY(puckposition);
        Point2D toCenter = Cartesian2D.vector(puPosNow, Cartesian2D.nullVector2D());
        toCenter = Cartesian2D.unitVector(toCenter);
        toCenter.x *= puckradius/2f;
        toCenter.y *= puckradius/2f;
        
        puPosNow = Cartesian2D.vectorSum(puPosNow, toCenter);
        setPuckPosition(Point2D.to3DmapXYtoXZ(puPosNow, puckposition[1]));
        
        if(puPosNow.length() <= puckradius) // puck near center?
        {
          setPuckPosition(0.0f, puckposition[1], 0.0f);
          puckState = 3;
        }
      }
      else if(puckState == 3 && gamecycle >=0) // puck descending (at center, after puck reached it)
      {
        puckposition[1] += puckAscDescSpeed;
        
        if(puckposition[1] >= 0.0f) // puck at table surface?
        {
          // do "who serves next" logic / center puck / accelerate puck towards that player
          setPuckPosition(0.0f, 0.0f, 0.0f);
          float m = whoServesNext == 0 ? -1.0f : 1.0f;
          setPuckSpeed(0.0f, 0.0f, m * puckServeSpeedZ);
          whoServesNext = (whoServesNext + 1)%2;
          
          // set puck to be in-game again
          puckState = 0;
          puckCollision = true;
        }
      }
    }
    
    // gather current positions and speeds for incremental collision detection/handling
    int p0 = 0;
    int p1 = 1;
    int pu = 2;
    Point2D pos[] = { Point2D.from3DmapXZtoXY(paddleposition[0]), 
                      Point2D.from3DmapXZtoXY(paddleposition[1]),
                      Point2D.from3DmapXZtoXY(puckposition)
                    };
    Point2D speed[] = { new Point2D(pdx[0], pdz[0]),
                        new Point2D(pdx[1], pdz[1]),
                        new Point2D(puckspeed[0], puckspeed[2])
                      };
    
    // sanitize paddle speed vectors
    Point2D prjTPos[] = { getSanitizedPaddleTargetXZPosition(0, pos[0], speed[0]),
                          getSanitizedPaddleTargetXZPosition(1, pos[1], speed[1]) };
    for(int i=0; i<2; i++)
    {
      if(prjTPos[i].x == prevPaddleposition[i].x)
        speed[i].x = 0f;
      if(prjTPos[i].y == prevPaddleposition[i].y)
        speed[i].y = 0f;
      
      prevPaddleposition[i] = prjTPos[i];
    }
    
    
   
    // find the fastest moving object to calculate the needed number of timesteps for incremental collision detection/handling
    float sRef = 0.0f;
    for(int i=0; i<speed.length; i++)
    {
      float s = speed[i].length();
      if(s > sRef)
      {
        sRef = s;
      }
    }
    
    if(sRef == 0f) // nothing moves == no new collisions possible
      return;
    
    // find the number of timesteps for incremental, limited precision collision tests
    // ( ts = fastest moving object displacement length(in cm) / desired precision(in cm) )
    int ts = (int) Math.ceil(sRef / 0.5f );
    
    // move all objects in small portions of their total speed and handle collisions
    // at each timestep
    // up/down/left/right in the following code refer to player ONE's point of view of the scene
    // or to be more precise to viewing the table from the top with z+ pointing north and x+ pointing east
    for(int t=0; t<ts; t++)
    {
      // calculate temporary puck destination and get border collision state for that
      Point2D relPKspeed = new Point2D(speed[pu].x / ts, speed[pu].y / ts);
      Point2D tPKpos = getSanitizedPuckTargetXZPosition(pos[pu], relPKspeed);
      int puckBorderColl = puckBorderCollision(tPKpos);
      
      // handle borders to puck collision
      if(puckBorderColl == 3 || puckBorderColl == 1)
      {
        speed[pu].x = -0.9f * speed[pu].x; // reverse and dampen a little
      }
      
      if(puckBorderColl == 0) // player 1 scores?
      {
        score[1]++;
        puckState = 1;
        puckCollision = false;
        break;
      }
      if(puckBorderColl == 2) // player 0 scores?
      {
        score[0]++;
        puckState = 1;
        puckCollision = false;
        break;
      }
      
      for(int i=0; i<2; i++) // move/handle paddles to puck collisions
      {
        // calculate temporary paddle destination and get puck collision state for that
        Point2D relSpeed = new Point2D(speed[i].x / ts, speed[i].y / ts);
        Point2D tPos = getSanitizedPaddleTargetXZPosition(i, pos[i], relSpeed);
        int puckPaddleColl = puckPaddleCollision(tPos, tPKpos);
        
        // handle puck to paddle collisions, adjust puck speed
        float reversez = 1.0f; // for changing puck z speed direction on collision
        float reversex = 1.0f; // for changing puck x speed direction on collision
        float dampenz = 1.0f; // for dampening puck z speed on collision
        float dampenx = 1.0f; // for dampening puxk x speed on collision
        float strongdampen = 0.2f;
        float weakdampen = 0.5f;
        float newsz = speed[pu].y; // new puck z speed on collision
        float newsx = speed[pu].x; // new puck x speed on collision
        if(puckPaddleColl == 0 || puckPaddleColl == 1 || puckPaddleColl == 2) // bottom of paddle
        {
          if(speed[pu].y > 0f) // puck moving up
          {
            if(speed[i].y == 0f) // paddle not moving in z
            {
              reversez = -1f;
              dampenz = weakdampen;
            }
            if(speed[i].y > 0f) // paddle moving up
            {
              reversez = -1f;
              dampenz = strongdampen;
            }
            if(speed[i].y < 0f) // paddle moving down
            {
              newsz = speed[i].y;
            }
          }
          if(speed[pu].y <= 0f) // puck not moving in z or moving down
          {
            if(speed[i].y < 0f) // paddle moving down
            {
              newsz = speed[i].y;
            }
          }
          
          if(puckPaddleColl == 1) // bottom only, no corner
          {
            newsx = (speed[pu].x + speed[i].x) / 2;
          }
        }
        if(puckPaddleColl == 4 || puckPaddleColl == 5 || puckPaddleColl == 6) // top of paddle
        {
          if(speed[pu].y >= 0f) // puck not moving in z or moving up
          {
            if(speed[i].y > 0f) // paddle moving up
            {
              newsz = speed[i].y;
            }
          }
          
          if(speed[pu].y < 0f) // puck moving down
          {
            if(speed[i].y == 0f) // paddle not moving in z
            {
              dampenz = weakdampen;
              reversez = -1f;
            }
            if(speed[i].y > 0f) // paddle moving up
            {
              newsz = speed[i].y;
            }
            if(speed[i].y < 0f) // paddle moving down
            {
              dampenz = strongdampen;
              reversez = -1f;
            }
          }
          
          if(puckPaddleColl == 5) // top only, no corner
          {
            newsx = (speed[pu].x + speed[i].x) / 2;
          }
        }
        
        if(puckPaddleColl == 6 || puckPaddleColl == 7 || puckPaddleColl == 0) // left of paddle
        {
          if(speed[pu].x > 0f) // puck moving right
          {
            if(speed[i].x == 0f) // paddle not moving in x
            {
              reversex = -1f;
              dampenx = weakdampen;
            }
            if(speed[i].x > 0f) // paddle moving right
            {
              reversex = -1f;
              dampenx = strongdampen;
            }
            if(speed[i].x < 0f) // paddle moving left
            {
              newsx = speed[i].x;
            }
          }
          if(speed[pu].x <= 0f) // puck not moving in x or moving left
          {
            if(speed[i].x < 0f) // paddle moving left
            {
              newsx = speed[i].x;
            }
          }
        }
        if(puckPaddleColl == 4 || puckPaddleColl == 3 || puckPaddleColl == 2) // right of paddle
        {
          if(speed[pu].x >= 0f) // puck not moving in x or moving right
          {
            if(speed[i].x > 0f) // paddle moving right
            {
              newsx = speed[i].x;
            }
          }
          
          if(speed[pu].x < 0f) // puck moving left
          {
            if(speed[i].x == 0f) // paddle not moving in x
            {
              dampenx = weakdampen;
              reversex = -1f;
            }
            if(speed[i].x > 0f) // paddle moving right
            {
              newsx = speed[i].x;
            }
            if(speed[i].x < 0f) // paddle moving left
            {
              dampenx = strongdampen;
              reversex = -1f;
            }
          }
        }
        
        // adjust puck z speed
        speed[pu].y = newsz * dampenz * reversez;
        // adjust puck x speed
        speed[pu].x = newsx * dampenx * reversex;
        
        if(puckPaddleColl != -1) // if there was any collision between puck and paddle
        {
          // recalculate next puck pos so it won't get stuck inside paddle
          relPKspeed = new Point2D(speed[pu].x / ts, speed[pu].y / ts);
          tPKpos = getSanitizedPuckTargetXZPosition(tPKpos, relPKspeed);
        }
        
        // handle puck blocking paddle movement at borders cases
        Point2D refCorner = null;
        if(puckBorderColl == 3 || puckPaddleColl == 0 || puckPaddleColl == 6) 
        {
          if(puckPaddleColl == 7) // puck block moving left
          {
            tPos.x = -tablewidth/2f + puckradius*2f + paddlewidth/2f;
            speed[i].x = 0f;
          }
          if(puckPaddleColl == 0) // puck may block lower left corner
            refCorner = new Point2D(tPos.x - paddlewidth/2f, tPos.y - paddledepth/2f);
          if(puckPaddleColl == 6) // puck may block upper left corner
            refCorner = new Point2D(tPos.x - paddlewidth/2f, tPos.y + paddledepth/2f);
        }
        if(puckBorderColl == 1 || puckPaddleColl == 2 || puckPaddleColl == 4)
        {
          if(puckPaddleColl == 3) // puck block moving right
          {
            tPos.x = tablewidth/2f - puckradius*2f - paddlewidth/2f;
            speed[i].x = 0f;
          }
          if(puckPaddleColl == 2) // puck may block lower right corner
            refCorner = new Point2D(tPos.x + paddlewidth/2f, tPos.y - paddledepth/2f);
          if(puckPaddleColl == 4) // puck may block upper right corner
            refCorner = new Point2D(tPos.x + paddlewidth/2f, tPos.y + paddledepth/2f);
        }
        if(refCorner != null)
        {
          Point2D pkToPaddle = Cartesian2D.vector(tPKpos, refCorner);
          float distToPuckCircle = puckradius - pkToPaddle.length();
          
          if(distToPuckCircle >= 0f) // if reference corner would be inside puck
          {
            speed[i].x = 0f;
            
            pkToPaddle = Cartesian2D.unitVector(pkToPaddle);
            pkToPaddle.x *= distToPuckCircle*1.0001;
            pkToPaddle.y *= distToPuckCircle*1.0001;
            tPos = Cartesian2D.vectorSum(tPos, pkToPaddle); // push paddle back to outer puck perimeter
          }
        }
        
        // adjust paddle position 
        pos[i] = tPos;
      }
      
      // adjust puck position
      pos[pu] = tPKpos;
    }
    
    // set new object positions and speeds
    puckposition = Point2D.to3DmapXYtoXZ(pos[pu], puckposition[1]);
    paddleposition[0] = Point2D.to3DmapXYtoXZ(pos[p0]);
    paddleposition[1] = Point2D.to3DmapXYtoXZ(pos[p1]);
    
    puckspeed = Point2D.to3DmapXYtoXZ(speed[pu]);
    setDesiredPaddleDisplacement(p0, 0.0f, 0.0f);
    setDesiredPaddleDisplacement(p1, 0.0f, 0.0f);
    
    // for TESTING/DEBUGGING:
    // displace puck by desired paddle displacement from player 0
    // setPuckPosition( puckposition[0] + pdx[0], 0f, puckposition[2] + pdz[0] );
  }

  /**
   * @return the game cycle
   */
  public int getGamecycle()
  {
    return gamecycle;
  }
  
  /**
   * @return the total number of game cycles per match
   */
  public int getTotalMatchCycles()
  {
    return this.totalMatchCycles;
  }
  
  /**
   * @return the number of pre match game cycles
   */
  public int getPreMatchCycles()
  {
    return this.preMatchCycles;
  }

  /**
   * calculates and returns the number of match seconds left in current match
   * (or before match reset if the returned value is negative)
   * @return 
   */
  public synchronized int getSecondsLeft()
  {
    if(gamecycle < 0)
    {
      return gamecycle / PaddlePuckGameState.CyclesPerSecond;
    }
    else
    {
      return (totalMatchCycles - gamecycle) / PaddlePuckGameState.CyclesPerSecond;
    }
  }
  
  /**
   * @return the paddle width
   */
  public float getPaddlewidth()
  {
    return paddlewidth;
  }

  /**
   * @return the paddle depth
   */
  public float getPaddledepth()
  {
    return paddledepth;
  }

  /**
   * @return the paddle height
   */
  public float getPaddleheight()
  {
    return paddleheight;
  }

  /**
   * @return the puck height
   */
  public float getPuckheight()
  {
    return puckheight;
  }

  /**
   * @return the puck radius
   */
  public float getPuckradius()
  {
    return puckradius;
  }

  /**
   * @return the puck position
   */
  public float[] getPuckPosition()
  {
    return puckposition.clone();
  }
  
  /**
   * simply sets the puck position to the given newPos
   * (does not check whether that new position is actually possible)
   * @param newPos new puck position as an array of three floats
   */
  public void setPuckPosition(float[] newPos)
  {
    if(newPos.length == 3)
      puckposition = newPos;
  }
  
  /**
   * simply sets the puck position to the given coordinates
   * (does not check whether that new position is actually possible)
   * @param x
   * @param y
   * @param z 
   */
  public void setPuckPosition(float x, float y, float z)
  {
    puckposition[0] = x;
    puckposition[1] = y;
    puckposition[2] = z;
  }
  
  /**
   * returns the puck speed vector
   * @return { sx, sy, sz } 
   */
  public float[] getPuckSpeed()
  {
    return puckspeed.clone();
  }
  
  /**
   * simply sets the puck speed to the given newSpeed
   * @param newSpeed new puck speed as an array of three floats
   */
  public void setPuckSpeed(float[] newSpeed)
  {
    if(newSpeed.length == 3)
      puckspeed = newSpeed;
  }
  
  /**
   * simply sets the puck speed to the given speed vector components
   * @param sx speed along x axis
   * @param sy speed along y axis
   * @param sz speed along z axis
   */
  public void setPuckSpeed(float sx, float sy, float sz)
  {
    puckspeed[0] = sx;
    puckspeed[1] = sy;
    puckspeed[2] = sz;
  }
  
  /**
   * @return the given clients paddle position
   */
  public float[] getPaddlePosition(int player)
  {
    if(player >= 0 && player < 2)
      return paddleposition[player].clone();
    else
      return new float[] { 0.0f, 0.0f, 0.0f };
  }
  
  /**
   * simply sets the paddle position for the given player to the given newPos
   * (does not check whether that new position is actually possible)
   * @param player 0 or 1
   * @param newPos new position as an array of three floats
   */
  public void setPaddlePosition(int player, float[] newPos)
  {
    if(player >=0 && player < 2 && newPos.length == 3)
      paddleposition[player] = newPos;
  }
  
  /**
   * simply sets the paddle position for the given player to the given coordinates
   * (does not check whether that new position is actually possible)
   * @param player 0 or 1
   * @param x
   * @param y
   * @param z 
   */
  public void setPaddlePosition(int player, float x, float y, float z)
  {
    if(player >= 0 && player < 2)
    {
      paddleposition[player][0] = x;
      paddleposition[player][1] = y;
      paddleposition[player][2] = z;
    }
  }
  
  /**
   * sets the desired paddle displacement for the given player
   * (will be used next time Update is called)
   * @param player 0 or 1
   * @param dx delta x to current position x
   * @param dz delta z to current position z
   */
  public void setDesiredPaddleDisplacement(int player, float dx, float dz)
  {
    if (player >=0 && player < 2)
    {
      pdx[player] = dx;
      pdz[player] = dz;
    }
  }
      
  /**
   * returns the score for the given player
   * @param player 0 or 1
   * @return 
   */
  public int getScore(int player)
  {
    if(player >= 0 && player < 2)
    {
      return score[player];
    }
    else
      return 0;
  }
  
  /**
   * sets the score for the given player to the given newScore
   * @param player
   * @param newScore 
   */
  public void setScore(int player, int newScore)
  {
    if(player >= 0 && player < 2)
    {
      score[player] = newScore;
    }
  }
  
  /**
   * @return the table width
   */
  public float getTablewidth()
  {
    return tablewidth;
  }

  /**
   * @return the table depth
   */
  public float getTabledepth()
  {
    return tabledepth;
  }

  /**
   * @return the home zone depth
   */
  public float getHomezonedepth()
  {
    return homezonedepth;
  }

  /**
   * @return the puckState
   */
  public int getPuckState()
  {
    return puckState;
  }

  /**
   * @param puckState the puckState to set
   */
  public void setPuckState(int puckState)
  {
    this.puckState = puckState;
  }
}
