/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckprotocol;

import util.Logging;
import util.SimpleRunnable;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.SocketTimeoutException;
import paddlepuckGUI.IPaddlePuckServerView;

/**
 * The Paddle Puck Server.
 * Runs a thread which listens for and accepts incoming connections.
 * That thread will also start/stop another thread for running the game logic as soon as
 * two clients are connected.
 */
public final class PaddlePuckServer
{
  private volatile boolean serverRunning = false;
  private Thread connectionInitiator = null;
  private Thread heartbeatsandgamestateupdates = null;
  private final int updatesPerSecond = PaddlePuckGameState.CyclesPerSecond;
  private volatile boolean cpuControlPaddle1 = false;
  private volatile boolean matchWasReset = false;
  
  private Logging logging = null;
  
  public static volatile int Version = 100;
  private int serverPort = 42000;
  private ServerSocket serverSocket = null;
  
  private final PaddlePuckConnectionServerSide[] client = new PaddlePuckConnectionServerSide[2];
  private final Integer[] clientLock = new Integer[] { new Integer(0), new Integer(1) };
  private final PaddlePuckGameState gamestate = new PaddlePuckGameState();
  
  private IPaddlePuckServerView ippsw = null;
  
  /**
   * Constructs PaddlePuckServer object (but does not start the server right away)
   * @param logging the logging object for logging messages
   * @param ip the IP on which to listen for incoming connection ("*" means any)
   * @param port the port on which to listen for incoming connections
   * @throws IOException exception as thrown by socket.bind
   * @throws Exception exception as thrown by SimpleRunnable.CreateThread
   */
  public PaddlePuckServer(Logging logging, String ip, int port) throws IOException, Exception
  {
    this.logging = logging;
    this.serverPort = port;
    
    serverSocket = new ServerSocket();
    serverSocket.setReuseAddress(true);
    if(ip.compareTo("*") == 0)
      serverSocket.bind(new InetSocketAddress(this.serverPort));
    else
    {
      serverSocket.bind(new InetSocketAddress(ip, this.serverPort));
    }
      
    serverSocket.setSoTimeout(1000);
    
    client[0] = null;
    client[1] = null;
    
    connectionInitiator = SimpleRunnable.CreateThread(this, "RunConnectionInitiatorThread", "connectionInitiator");
    heartbeatsandgamestateupdates = SimpleRunnable.CreateThread(this, "RunHeartbeatsAndGameStateUpdatesThread", "heartbeats");
  }

  /**
   * sets the view which will get gamestate-changed signals
   * (only works while the server is not running)
   * @param ippsw the view to signal on gamestate-changes
   */
  public void setView(IPaddlePuckServerView ippsw)
  {
    if(!serverRunning)
      this.ippsw = ippsw;
  }
  
  /**
   * sets the CPUcontrolPaddle1 flag
   * @param newValue set this to true to have paddle 1 (Player TWO) controlled by the CPU
   */
  public void setCPUcontrolPaddle1(boolean newValue)
  {
    this.cpuControlPaddle1 = newValue;
  }
  
  /**
   * get the game state
   */
  public PaddlePuckGameState getGameState()
  {
    return this.gamestate;
  }
  
  /**
   * Starts the server.
   */
  public void StartServer()
  {
    serverRunning = true;
    
    connectionInitiator.start();
    heartbeatsandgamestateupdates.start();
    
    this.logging.LogMessage("Server started.");
  }
  
  /**
   * Stops the server.
   */
  public void StopServer()
  {
    serverRunning = false;
    
    synchronized(client)
    {
      for(int i=0; i<2; i++)
        if(client[i] != null)
        {
          try
          {
            // send message to client to close connection on their side
            client[i].SendCloseConnection();

            client[i].Close();
            client[i].StopProcessingClientMessages(true);
          }
          catch (Exception e)
          {
            this.logging.LogException(e);
          }
          finally
          {
            client[i] = null;
          }
        }
    }
    
    // wait for the heartbeats thread
    this.logging.LogMessage("joining heartbeats and gamestate updates thread...");
    try
    {
      heartbeatsandgamestateupdates.join();
    }
    catch (InterruptedException ex)
    {
      this.logging.LogException(ex);
    }
    
    // wait for the connectionInitiator thread
    this.logging.LogMessage("joining connectionInitiator thread...");
    try
    {
      connectionInitiator.join();
    }
    catch(InterruptedException ie)
    {
      this.logging.LogException(ie);
    }
    
    // close server socket
    try
    {
      this.logging.LogMessage("closing server socket...");
      this.serverSocket.close();
    }
    catch (Exception e)
    {
      this.logging.LogException(e);
    }
    
    this.logging.LogMessage("Server stopped.");
  }
  
  /**
   * sets the matchWasReset flag
   * (if it's set to true, then the next Update will send scores/time update messages to clients)
   * @param scoresReset
   */
  public void setMatchWasReset(boolean matchWasReset)
  {
    this.matchWasReset = matchWasReset;
  }
  
  /**
   * heartbeats and game updates thread
   * sends heartbeats every now and then to connected clients, so server gets messages from them even if they don't do anything (like generating input)
   * also updates the game state and sends game state position updates to connected clients
   */
  public void RunHeartbeatsAndGameStateUpdatesThread()
  {  
    int count = 0;
    
    while(serverRunning)
    {
      try
      {
        // for pre and post update states
        int psPre = 0;
        int psPost = 0;
        int[] sPre = new int[2];
        int[] sPost = new int[2];
        int preSecs = 0;
        int postSecs = 0;
        
        synchronized(gamestate)
        {
          // get pre-update states
          psPre = gamestate.getPuckState();
          sPre[0] = gamestate.getScore(0);
          sPre[1] = gamestate.getScore(1);
          preSecs = gamestate.getSecondsLeft();
          
          gamestate.Update(cpuControlPaddle1);
          
          // get post update states
          psPost = gamestate.getPuckState();
          sPost[0] = gamestate.getScore(0);
          sPost[1] = gamestate.getScore(1);
          postSecs = gamestate.getSecondsLeft();
          
          count = (count + 1) % (updatesPerSecond / 2);
        }
        
        // signal view (if there is one)
        if(this.ippsw != null)
          this.ippsw.SignalGameStateChanged(gamestate);
        
        for(int i=0; i<2; i++) // send update message(s) to clients
        {
          if(client[i] != null)
          {
            synchronized(clientLock[i])
            {
              if(count == 0)
                client[i].SendHeartbeat(); // client will respond to this

              client[i].SendGameStatePositionsUpdate(gamestate);

              if(psPre != psPost)
                client[i].SendGameStatePuckStateUpdate(gamestate);
              
              if(preSecs != postSecs)
                client[i].SendGameStateMatchSecondsLeft(gamestate);

              if(sPre[0] != sPost[0] || sPre[1] != sPost[1] || matchWasReset)
              {
                client[i].SendGameStateScoresUpdate(gamestate);
                client[i].SendGameStateMatchSecondsLeft(gamestate);
                client[i].SendGameStatePuckStateUpdate(gamestate);
              }
            }
          }
        }
        
        // clear flags
        matchWasReset = false;
      }
      catch (Exception e)
      {
        this.logging.LogException(e);
      }
      
      try
      {
        Thread.sleep(1000 / updatesPerSecond);
      }
      catch (InterruptedException ex)
      {
        this.logging.LogException(ex);
      }
    }
  }
  
  /*
   * connectionInitiator thread
   * accepts incoming connections and starts processing client messages received in the
   * PaddlePuckConnectionServerSide objects own threads for that
   */
  public void RunConnectionInitiatorThread()
  {
    while(serverRunning)
    {
      Thread.yield();
      
      try
      {
        boolean acceptDummy = false;
        
        // determine connection states before accepting new connections
        boolean preC0con = client[0] != null;
        boolean preC1con = client[1] != null;

        // accept player connections
        for(int i=0; i<2; i++)
        {
          if(client[i] == null)
          {
            synchronized(clientLock[i])
            {
              client[i] = new PaddlePuckConnectionServerSide(this, serverSocket.accept(), Version, i, this.logging);
              client[i].StartProcessingClientMessages();
              matchWasReset = true; // so connected player will receive current scores with first update
            }
            this.logging.LogMessage(String.format("player %1$s connected.", i == 0 ? "ONE":"TWO"));
          }
        }
          
        // determine connection states after accepting new connections
        boolean postC0con = client[0] != null;
        boolean postC1con = client[1] != null;
        
        // compare with previous connection states 
        // initialize game if both players are connected now (and if at least one wasn't before)         
        if(postC0con && postC1con && (!preC0con || !preC1con))
        {
          synchronized(this.gamestate)
          {
            this.gamestate.Reset();
            matchWasReset = true;
            
            // TODO: send player connected messages
          }
        }
        
        acceptDummy = (client[0] != null && client[1] != null);
        
        if(acceptDummy)
        {
          // accept other connections (only to tell them that the server does not allow any more players (playerNumber -1))
          PaddlePuckConnectionServerSide dummy = new PaddlePuckConnectionServerSide(this, serverSocket.accept(), Version, -1, this.logging);
          dummy = null;
        }
      }
      catch(SocketTimeoutException e)
      {
        this.logging.LogMessage("waiting for connections...");
      }
      catch(Exception e)
      {
        this.logging.LogException(e); 
      }
    }
  }
  
  /**
   * processes the CloseConnection message received in a PaddlePuckConnectionServerSide processIncomingMessages thread
   * @param playerNumber 
   */
  public void HandleClientDisconnectedByMessage(int playerNumber)
  {
    if(playerNumber >=0 && playerNumber <2)
    {
      synchronized(client)
      {
        try
        {
          if(client[playerNumber] != null)
          {
            client[playerNumber].Close();
            client[playerNumber].StopProcessingClientMessages(false);
            this.logging.LogMessage(String.format("Player %1$s disconnected.", playerNumber == 0 ? "ONE":"TWO"));
          }
        }
        catch(Exception e)
        {
          this.logging.LogException(e);
        }
        finally
        {
          client[playerNumber] = null;
        }
      }
    }
  }
  
  /**
   * processes the SocketTimeoutException in a PaddlePuckConnectionServerSide processIncomingMessages thread
   * @param playerNumber 
   */
  public void HandleClientDisconnectedByTimeout(int playerNumber)
  {
    this.logging.LogMessage(String.format("Disconnecting player %1$s due to SocketTimeoutException...", playerNumber == 0 ? "ONE":"TWO"));
    this.HandleClientDisconnectedByMessage(playerNumber);
  }
  
  /**
   * processes the EOFException in a PaddlePuckConnectionServerSide processIncomingMessages thread
   * @param playerNumber 
   */
  public void HandleClientDisconnectedByEndOfStream(int playerNumber)
  {
    this.logging.LogMessage(String.format("Disconnecting player %1$s due to EOFException on incoming data stream...", playerNumber == 0 ? "ONE":"TWO"));
    this.HandleClientDisconnectedByMessage(playerNumber);
  }
  
  /**
   * processes the IOException in a PaddlePuckConnectionServerSide processIncomingMessages thread
   * @param playerNumber 
   */
  public void HandleClientDisconnectedByIOError(int playerNumber)
  {
    this.logging.LogMessage(String.format("Disconnecting player %1$s due to IOExcpetion on incoming data stream...", playerNumber));
    this.HandleClientDisconnectedByMessage(playerNumber);
  }
  
  /**
   * processes the DesiredPaddleDisplacementMessage
   * @param playerNumber which paddle to displace
   * @param dx the delta x as received in the message data
   * @param dz the delta y as received in the message data
   */
  public void HandleDesiredPaddleDisplacementMessage(int playerNumber, float dx, float dz)
  {
    synchronized(gamestate)
    {
      // each player sees the table from a different side but input is grabbed the same way
      // on the client side, regardless of view
      // so the sign gets switched around here, so that the movement in the gamestate coordinate
      // system is in the correct direction
      gamestate.setDesiredPaddleDisplacement(playerNumber, playerNumber == 0 ? dx : -dx, playerNumber == 0 ? dz : -dz);
    }
  }
}
