/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package util.MyMath;

/**
 * my 2D "cartesian coordinate system 
 * math routines" ported from C++ to Java on Oct. 15th 2011
 * 
 * (illustration of axes and angles for reference with method comments)
 * 0,0---1--..x+
 *  | \cw | 
 *  |  \ /        cw = clockwise angle of the vector p in DEG
 *  |   \         (the counter clockwise one is 360 DEG - cw)
 *  1    p(1,1)
 *  | 
 *  | 
 *  ..
 *  y+
 */
public class Cartesian2D
{
  public static final float myPI = 3.14159265358979f; // PI as a constant
  public static final float toDegFactor = 180.0f / myPI; // multiply with an angle in RAD to get DEG
  public static final float toRadFactor = myPI / 180.0f; // multiply with an angle in DEG to get RAD
  
  /**
   * @return the 2D 0-vector
   */
  public static Point2D nullVector2D() { return new Point2D(); };
  
  /**
   * @param pA point A
   * @param pB point B
   * @return pythagorean distance between two points
   */
  public static float distance(Point2D pA, Point2D pB)
  {
    return vector(pA, pB).length();
  }

 
  /**
   * @param pA point A
   * @param pB point B
   * @return vector from pA to pB
   */
  public static Point2D vector(Point2D pA, Point2D pB)
  {
    return new Point2D(pB.x - pA.x, pB.y - pA.y);
  }

  /**
   * @param pA a point/vector
   * @return unit vector of pA
   */
  public static Point2D unitVector(Point2D pA)
  {
    float l = pA.length();
    if (l != 0f)
      return new Point2D(pA.x / l, pA.y / l);
    else
      return nullVector2D();
  }

  /**
   * @param pA a vector
   * @return length of pA
   */
  public static float vectorLength(Point2D pA)
  {
    return pA.length();
  }

  /**
   * @param pA a point
   * @return angle between (1,0) and vector from (0,0) to pA (clockwise)
   */
  public static float vectorAngleCW(Point2D pA)
  {
    if(pA.x == 0.0f && pA.y == 0.0f)
      return 0.0f;
    else
    {
      float len = pA.length();
      float angle = (float)Math.acos(pA.x / len) * toDegFactor;
      if(pA.y > 0.0f)
      {
        return angle;
      }
      else if(pA.y < 0.0f)
      {
        return 360.0f - angle;
      }
      else
        return angle; // y == 0.0f
    }
  }

  /**
   * @param pA a point
   * @return angle between (1,0) and vector from (0,0) to pA (counterclockwise)
   */
  public static float vectorAngleCCW(Point2D pA)
  {
    if(pA.x == 0.0f && pA.y == 0.0f)
      return 0.0f;
    else
    {
      float len = pA.length();
      float angle = (float)Math.acos(pA.x / len) * toDegFactor;
      if(pA.y < 0.0f)
      {
        return angle;
      }
      else if(pA.y > 0.0f)
      {
        return 360.0f - angle;
      }
      else
        return angle; // y == 0.0f
    }
  }

  /**
   * construct a vector from clockwise polar coordinates
   * @param length the length of the vector to construct
   * @param angleDeg the clockwise angle in degrees between the x axis and the vector to construct
   * @return vector constructed from arguments e.g.: (1, 90°) -> (0, 1)
   */
  public static Point2D vectorFromPolarCW(float length, float angleDeg)
  {
    float angRad = angleDeg * toRadFactor;
    Point2D v = new Point2D((float)Math.cos(angRad) * length, (float)Math.sin(angRad) * length);
    return v;
  }

  /**
   * construct a vector from counterclockwise polar coordinates
   * @param length the length of the vector to construct
   * @param angleDeg the counterclockwise angle in degrees between the x axis and the vector to construct
   * @return vector constructed from arguments e.g.: (1, 90°) -> (0,-1)
   */
  public static Point2D vectorFromPolarCCW(float length, float angleDeg)
  {
    float angRad = angleDeg * toRadFactor;
    Point2D v = new Point2D((float)Math.cos(angRad) * length, -1f * (float)Math.sin(angRad) * length);
    return v;
  }

  /**
   * add two vectors
   * @param pA vector A
   * @param pB vector B
   * @return the sum of the given vectors
   */
  public static Point2D vectorSum(Point2D pA, Point2D pB)
  {
    return new Point2D(pA.x + pB.x, pA.y + pB.y);
  }

  /**
   * normalize an angle, so it's always between 0.0f and 360.0f
   * @param angle the angle to normalize
   * @return the normalized angle
   */
  public static float normalizeAngle(float angle)
  {
    float tmpAngle = angle%360f;
    if (tmpAngle < 0f)
      tmpAngle = -tmpAngle;
    return tmpAngle;
  }
}
