/* Project: Paddle Puck
 * Author: Dennis Busch, http://www.dennisbusch.de, Dennis[minus]Busch[at]gmx[dot]net
 */

package paddlepuckserver;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;
import paddlepuckGUI.GUIComponentHandlers;
import paddlepuckGUI.IPaddlePuckServerView;
import paddlepuckprotocol.PaddlePuckGameState;
import util.Logging;
import paddlepuckprotocol.PaddlePuckServer;
import util.MyMath.Cartesian2D;
import util.MyMath.Point2D;

/**
 * The front-end for the PaddlePuckServer.
 */
public class PaddlePuckServerGUI extends java.awt.Frame implements IPaddlePuckServerView
{
  private PaddlePuckServer server = null;
  private Logging logging = null;
  private volatile boolean skipPaint = false;
  
  // backbuffer for flicker-free canvas updates
  Image backbuffer = null;
  
  /** Creates new form PaddlePuckServerGUI */
  public PaddlePuckServerGUI()
  {
    initComponents();
    
    // set title and init logging
    this.setTitle(String.format("Paddle Puck Server (%1$d)", PaddlePuckServer.Version));
    this.logging = new Logging(this.listLog);
    
    // init backbuffer for gamestate view
    backbuffer = this.canvasView.createImage(this.canvasView.getWidth(), this.canvasView.getHeight());
    
    // determine/add network adapters to dropdown list
    this.ddlServerAdapter.add("*");
    
    try
    {
      Enumeration<NetworkInterface> nis = NetworkInterface.getNetworkInterfaces();
      
      while(nis.hasMoreElements())
      {
        NetworkInterface ni = nis.nextElement();
        Enumeration<InetAddress> ias = ni.getInetAddresses();
        while(ias.hasMoreElements())
        {
          InetAddress ia = ias.nextElement();
          this.ddlServerAdapter.add(ia.getHostAddress());
        }
      }
    }
    catch(SocketException se)
    {
      this.logging.LogMessage("Exception occured while enumerating network interfaces.");
      this.logging.LogException(se);
    }
    finally
    {
      //if(this.ddlServerAdapter.getItemCount() >= 2)
      //{
      //  this.ddlServerAdapter.select(1); // first determined IP address
      //}
      //else
      this.ddlServerAdapter.select(0); // "*" wildcard address
    }
  }
  
  /**
   * instantiates and starts the server object (parsing the textFieldPort for the port to use)
   * and updates GUI state
   */
  private void StartSever()
  {
    try
    {
      int port = Integer.parseInt(textFieldPort.getText());
      
      this.server = new PaddlePuckServer(this.logging, this.ddlServerAdapter.getSelectedItem(), port);
      
      this.skipPaint = false;
      this.server.setView((IPaddlePuckServerView)this);
      
      this.server.StartServer();
      
      this.setGUIstate(true);
    }
    catch (Exception e)
    {
      this.logging.LogException(e);
    }
  }
  
  /**
   * stops the server object, sets it to null
   * and updates GUI state
   */
  private void StopServer()
  {
    try
    {
      if(server != null)
      {
        server.StopServer();
        server = null;
        
        this.setGUIstate(false);
      }
    }
    catch (Exception e)
    {
      this.logging.LogException(e);
    }
  }

  /** This method is called from within the constructor to
   * initialize the form.
   * WARNING: Do NOT modify this code. The content of this method is
   * always regenerated by the Form Editor.
   */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        listLog = new java.awt.List();
        textFieldPort = new java.awt.TextField();
        labelPort = new java.awt.Label();
        buttonStartServer = new java.awt.Button();
        buttonStopServer = new java.awt.Button();
        btnResetGameState = new java.awt.Button();
        cbCPUcontrol = new java.awt.Checkbox();
        ddlServerAdapter = new java.awt.Choice();
        lblNetworkInterface = new java.awt.Label();
        canvasView = new java.awt.Canvas();

        setLocationRelativeTo(null);
        setMinimumSize(new java.awt.Dimension(640, 480));
        setName("framePaddlePuckGUI"); // NOI18N
        setResizable(false);
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                exitForm(evt);
            }
        });

        listLog.setName("listLog"); // NOI18N

        textFieldPort.setName("textFieldPort"); // NOI18N
        textFieldPort.setText("42000");
        textFieldPort.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyTyped(java.awt.event.KeyEvent evt) {
                textFieldPortKeyTyped(evt);
            }
        });

        labelPort.setName("labelPort"); // NOI18N
        labelPort.setText("Port:");

        buttonStartServer.setLabel("Start Server");
        buttonStartServer.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonStartServerActionPerformed(evt);
            }
        });

        buttonStopServer.setEnabled(false);
        buttonStopServer.setLabel("Stop Server");
        buttonStopServer.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonStopServerActionPerformed(evt);
            }
        });

        btnResetGameState.setLabel("Reset GameState");
        btnResetGameState.setName("btnResetGameState"); // NOI18N
        btnResetGameState.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btnResetGameStateActionPerformed(evt);
            }
        });

        cbCPUcontrol.setLabel("CPU controls Player TWO");
        cbCPUcontrol.setName("cbCPUcontrol"); // NOI18N
        cbCPUcontrol.addItemListener(new java.awt.event.ItemListener() {
            public void itemStateChanged(java.awt.event.ItemEvent evt) {
                cbCPUcontrolItemStateChanged(evt);
            }
        });

        ddlServerAdapter.setName("ddlServerAdapter"); // NOI18N

        lblNetworkInterface.setName("lblNetworkInterface"); // NOI18N
        lblNetworkInterface.setText("Network Interface:");

        canvasView.setBackground(java.awt.Color.black);
        canvasView.setName("canvasView"); // NOI18N

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(listLog, javax.swing.GroupLayout.DEFAULT_SIZE, 687, Short.MAX_VALUE)
                        .addContainerGap())
                    .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                            .addGroup(layout.createSequentialGroup()
                                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                                    .addComponent(cbCPUcontrol, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 254, Short.MAX_VALUE)
                                    .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING, false)
                                        .addComponent(btnResetGameState, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                                        .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createSequentialGroup()
                                            .addComponent(buttonStartServer, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                                            .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                            .addComponent(buttonStopServer, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))))
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 48, Short.MAX_VALUE))
                            .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createSequentialGroup()
                                .addComponent(lblNetworkInterface, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                .addComponent(ddlServerAdapter, javax.swing.GroupLayout.PREFERRED_SIZE, 190, javax.swing.GroupLayout.PREFERRED_SIZE)))
                        .addGap(49, 49, 49)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING, false)
                            .addGroup(layout.createSequentialGroup()
                                .addComponent(labelPort, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                .addComponent(textFieldPort, javax.swing.GroupLayout.PREFERRED_SIZE, 72, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addGap(235, 235, 235))
                            .addGroup(layout.createSequentialGroup()
                                .addComponent(canvasView, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                                .addContainerGap())))))
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(listLog, javax.swing.GroupLayout.PREFERRED_SIZE, 246, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(lblNetworkInterface, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(ddlServerAdapter, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(textFieldPort, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(labelPort, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(layout.createSequentialGroup()
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(buttonStartServer, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(buttonStopServer, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(btnResetGameState, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(cbCPUcontrol, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addGap(96, 96, 96))
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(canvasView, javax.swing.GroupLayout.DEFAULT_SIZE, 174, Short.MAX_VALUE)
                        .addContainerGap())))
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

  /** Exit the Application */
    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm
      this.StopServer();
      System.exit(0);
    }//GEN-LAST:event_exitForm

  private void textFieldPortKeyTyped(java.awt.event.KeyEvent evt)//GEN-FIRST:event_textFieldPortKeyTyped
  {//GEN-HEADEREND:event_textFieldPortKeyTyped
    GUIComponentHandlers.PortTextFieldKeyTyped(evt);
  }//GEN-LAST:event_textFieldPortKeyTyped

  private void buttonStartServerActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_buttonStartServerActionPerformed
  {//GEN-HEADEREND:event_buttonStartServerActionPerformed
    this.StartSever();
    if(this.server != null)
      this.server.setCPUcontrolPaddle1(this.cbCPUcontrol.getState());
  }//GEN-LAST:event_buttonStartServerActionPerformed

  private void buttonStopServerActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_buttonStopServerActionPerformed
  {//GEN-HEADEREND:event_buttonStopServerActionPerformed
    this.StopServer();
  }//GEN-LAST:event_buttonStopServerActionPerformed

  private void btnResetGameStateActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_btnResetGameStateActionPerformed
  {//GEN-HEADEREND:event_btnResetGameStateActionPerformed
    if(this.server != null)
    {
      this.server.getGameState().Reset();
      this.server.setMatchWasReset(true);
    }
  }//GEN-LAST:event_btnResetGameStateActionPerformed

  private void cbCPUcontrolItemStateChanged(java.awt.event.ItemEvent evt)//GEN-FIRST:event_cbCPUcontrolItemStateChanged
  {//GEN-HEADEREND:event_cbCPUcontrolItemStateChanged
    if(this.server != null)
      this.server.setCPUcontrolPaddle1(this.cbCPUcontrol.getState());
  }//GEN-LAST:event_cbCPUcontrolItemStateChanged
   
  /**
   * used to enable/disable GUI elements according to server state
   * @param running 
   */
  private void setGUIstate(boolean running)
  {
    if(running)
    {
      textFieldPort.setEnabled(false);
      buttonStartServer.setEnabled(false);
      buttonStopServer.setEnabled(true);
      ddlServerAdapter.setEnabled(false);
    }
    else
    {
      textFieldPort.setEnabled(true);
      buttonStartServer.setEnabled(true);
      buttonStopServer.setEnabled(false);
      ddlServerAdapter.setEnabled(true);
    }
  }
  
  /**
   * implementation of IPaddlePuckServerView
   * is called asynchronously from the PaddlePuckServer updates thread
   * to signal game-state changes
   * @param gs the game-state
   */
  @Override
  public void SignalGameStateChanged(PaddlePuckGameState gs)
  {
    if(skipPaint)
      return;
    
    skipPaint = true;
    
    Point2D p0 = null; // paddle 0 pos
    Point2D p1 = null; // paddle 1 pos
    Point2D pu = null; // puck pos
    
    float pw = 0f; // paddle width
    float pd = 0f; // paddle depth
    
    float pr = 0f; // puck radius
    float tw = 0f; // table width
    float td = 0f; // table depth
    float hzd = 0f; // homezonedepth
    
    int s0 = 0; // score 0
    int s1 = 0; // score 1
    int ps = 0; // puck state
    
    int seconds = 0;
    
    // scaling
    Point2D cCanv = null; // canvas center
    float unit = 0f;
    
    // get information from gamestate
    synchronized(gs)
    {
      p0 = Point2D.from3DmapXZtoYX(gs.getPaddlePosition(0));
      p1 = Point2D.from3DmapXZtoYX(gs.getPaddlePosition(1));
      pu = Point2D.from3DmapXZtoYX(gs.getPuckPosition());
      
      pw = gs.getPaddlewidth();
      pd = gs.getPaddledepth();
      pr = gs.getPuckradius();
      tw = gs.getTablewidth();
      td = gs.getTabledepth();
      hzd = gs.getHomezonedepth();
      s0 = gs.getScore(0);
      s1 = gs.getScore(1);
      
      ps = gs.getPuckState();
      
      seconds = gs.getSecondsLeft();
    }
    
    // set up scaling and canvas center
    cCanv = new Point2D(this.canvasView.getWidth() / 2f, this.canvasView.getHeight() / 2f);
    unit = this.canvasView.getWidth() / td;
    pw *= unit;
    pd *= unit;
    pr *= unit;
    tw *= unit;
    td *= unit;
    hzd *= unit;
    p0.scale(unit);
    p1.scale(unit);
    pu.scale(unit);
    
    // get scaled/translated paddle and puck positions
    p0 = Cartesian2D.vectorSum(cCanv, p0);
    p1 = Cartesian2D.vectorSum(cCanv, p1);
    pu = Cartesian2D.vectorSum(cCanv, pu);
    
    // draw gamestate on backbuffer
    Graphics gFront = this.canvasView.getGraphics();    
    Graphics g = backbuffer.getGraphics();
    
    // clear
    g.clearRect(0, 0, this.canvasView.getWidth(), this.canvasView.getHeight());
    
    // draw table main piece
    g.setColor(Color.ORANGE);
    g.fillRect((int)(cCanv.x - td/2 + hzd), (int)(cCanv.y - tw/2), (int)(td - 2f*hzd), (int)tw);
    
    // draw home zones
    g.setColor(Color.GREEN);
    g.fillRect((int)(cCanv.x - td/2), (int)(cCanv.y - tw/2), (int)hzd, (int)tw);
    g.setColor(Color.CYAN);
    g.fillRect((int)(cCanv.x + td/2 - hzd), (int)(cCanv.y - tw/2), (int)hzd, (int)tw);
    
    // draw puck
    if(ps == 0)
      g.setColor(Color.DARK_GRAY);
    else
      g.setColor(Color.RED);
    g.fillOval((int)(pu.x - pr), (int)(pu.y - pr), (int)(pr*2f), (int)(pr*2f));
    
    // draw paddles
    g.setColor(Color.BLUE);
    g.fillRect((int)(p0.x - pd/2f), (int)(p0.y - pw/2f), (int)pd, (int)pw);
    g.fillRect((int)(p1.x - pd/2f), (int)(p1.y - pw/2f), (int)pd, (int)pw);
    
    // draw scores
    g.setColor(Color.GREEN); 
    g.drawString(String.format("Player ONE: %1$d", s0), 0, 10);
    
    g.setColor(Color.CYAN);
    g.drawString(String.format("Player TWO: %1$d", s1), 0, 20);
    
    // draw match time or pre match countdown
    g.setColor(Color.WHITE);
    if(seconds < 0)
    {
      g.drawString(String.format("next match starts in %1$ds", -(seconds)), 150, 10);
    }
    else
    {
      g.drawString(String.format("%1$02d:%2$02d", seconds / 60, seconds%60), 150, 10);
    }
    
    // draw backbuffer on canvas
    gFront.drawImage(backbuffer, 0, 0, this.canvasView);
    
    skipPaint = false;
  }
  
  /**
   * @param args the command line arguments
   */
  public static void main(String args[])
  {
    java.awt.EventQueue.invokeLater(new Runnable()
    {
      @Override
      public void run()
      {
        new PaddlePuckServerGUI().setVisible(true);
      }
    });
  }
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private java.awt.Button btnResetGameState;
    private java.awt.Button buttonStartServer;
    private java.awt.Button buttonStopServer;
    private java.awt.Canvas canvasView;
    private java.awt.Checkbox cbCPUcontrol;
    private java.awt.Choice ddlServerAdapter;
    private java.awt.Label labelPort;
    private java.awt.Label lblNetworkInterface;
    private java.awt.List listLog;
    private java.awt.TextField textFieldPort;
    // End of variables declaration//GEN-END:variables
}
